# -*- coding: utf-8 -*-

# Imports.
# Local imports.
from etacad.drawing_utils import mtext, rect
from etacad.globals import CADTABLE_SET_DEFAULT, Aligment, ElementTypes
from etacad.utils import max_per_position, text_width_estimation

# External imports.
from attrs import define, field
from ezdxf.document import Drawing


@define
class CADTable:
    data: list[list] = field(default=None)
    rows: int = field(default=None)
    columns: int = field(default=None)
    labels: list = field(default=None)
    index: list = field(default=None)
    x: float = field(default=0)
    y: float = field(default=0)
    settings: dict = field(default=CADTABLE_SET_DEFAULT)
    element_type: ElementTypes = ElementTypes.CADTABLE

    # Table dimensions.
    row_heights: list = field(init=False)
    column_widths: list = field(init=False)
    rows_length: float = field(init=False)
    columns_height: float = field(init=False)
    column_width_labels: list = field(init=False)

    def __attrs_post_init__(self):
        # Table dimensions content.
        if self.settings["content_fit"] and self.data:
            self.rows = len(self.data)
            self.columns = len(self.data[0])
            self.row_heights = [self.settings["content_row_height"]] * self.rows

            column_widths = [0] * self.columns
            for row in self.data:
                for i, content in enumerate(row):
                    width = text_width_estimation(content, self.settings["content_text_height"])
                    if width > column_widths[i]:
                        column_widths[i] = width

            self.column_widths = column_widths

        elif self.data:
            self.rows = len(self.data)
            self.columns = len(self.data[0])
            self.row_heights = [self.settings["content_row_height"]] * self.rows
            self.column_widths = [self.settings["content_row_height"]] * self.columns

        else:
            self.row_heights = [self.settings["content_row_height"]] * self.rows
            self.column_widths = [self.settings["content_row_height"]] * self.columns

        self.rows_length = sum(self.column_widths)
        self.columns_height = sum(self.row_heights)

        # Table dimensions labels.
        if self.settings["labels_fit"]:
            self.column_width_labels = [text_width_estimation(label, self.settings["labels_text_height"]) for label in self.labels]
        else:
            self.column_width_labels = self.settings["labels_column_width"]

    def cell_content_edit(self, row: int, column: int, text=str):
        self.data[row][column] = text

    def draw_grid(self, document: Drawing,
                  rows: int = None,
                  columns: int = None,
                  x: float = None,
                  y: float = None,
                  row_heights: list = None,
                  column_widths: list = None) -> dict:
        if rows is None:
            rows = self.rows

        if columns is None:
            columns = self.columns

        if x is None:
            x = self.x

        if y is None:
            y = self.y

        if row_heights is None:
            row_heights = self.row_heights

        if column_widths is None:
            column_widths = self.column_widths

        rows_length, columns_height = sum(column_widths), sum(row_heights)

        grid_hz_lines = []
        y_hz_line = y
        for i in range(rows):
            sides = [0, 0, 1, 0] if i < rows - 1 else [1, 0, 1, 0]
            grid_hz_lines += rect(doc=document,
                                  width=rows_length,
                                  height=row_heights[i],
                                  x=x,
                                  y=y_hz_line,
                                  sides=sides)
            y_hz_line += row_heights[i]

        grid_vt_lines = []
        x_vt_lines = x
        for i in range(columns):
            sides = [0, 0, 0, 1] if i < columns - 1 else [0, 1, 0, 1]
            grid_vt_lines += rect(doc=document,
                                  width=column_widths[i],
                                  height=columns_height,
                                  x=x_vt_lines,
                                  y=y,
                                  sides=sides)
            x_vt_lines += column_widths[i]

        # Setting groups of elements in dictionary.
        elements = {"grid_hz_lines": grid_hz_lines,
                    "grid_vt_lines": grid_vt_lines,
                    "all_elements": grid_hz_lines + grid_vt_lines}

        return elements

    def draw_labels(self, document: Drawing,
                    x: float = None,
                    y: float = None,
                    row_height: float = None,
                    column_widths: list = None) -> dict:
        if x is None:
            x = self.x

        if y is None:
            y = self.y + self.columns_height

        if row_height is None:
            row_height = [self.settings["labels_row_height"]]

        if column_widths is None:
            column_widths = self.column_widths

        elements = self.draw_grid(document=document,
                                  x=x,
                                  y=y,
                                  rows=1,
                                  columns=len(column_widths),
                                  row_heights=row_height,
                                  column_widths=column_widths)

        texts, x_mtext = [], x
        for i, label in enumerate(self.labels):
            x_mtext += column_widths[i] / 2
            texts += mtext(document=document,
                           textstr=label,
                           height=self.settings["labels_text_height"],
                           x=x_mtext,
                           y=y + self.settings["labels_row_height"] / 2,
                           width=column_widths[i],
                           rotation=0,
                           aligment=Aligment.MTEXT_MIDDLE_CENTER.value)
            x_mtext += column_widths[i] / 2
        elements["texts"] = texts

        # Setting groups of elements in dictionary.
        elements["all_elements"] = elements["all_elements"] + elements["texts"]

        return elements

    def draw_content(self, document: Drawing,
                     x: float = None,
                     y: float = None,
                     row_height: float = None,
                     column_widths: list = None) -> dict:
        if x is None:
            x = self.x

        if y is None:
            y = self.y + self.columns_height

        if row_height is None:
            row_height = self.row_heights

        if column_widths is None:
            column_widths = self.column_widths

        elements = self.draw_grid(document=document,
                                  x=x,
                                  y=y,
                                  rows=self.rows,
                                  columns=self.columns,
                                  row_heights=row_height,
                                  column_widths=column_widths)

        y_mtext = y
        reversed_data = self.data.copy()
        reversed_data.reverse()
        for j, row in enumerate(reversed_data):
            y_mtext += row_height[j] / 2
            texts, x_mtext = [], x
            for i, text in enumerate(row):
                x_mtext += column_widths[i] / 2
                texts += mtext(document=document,
                               textstr=text,
                               height=self.settings["content_text_height"],
                               x=x_mtext,
                               y=y_mtext,
                               width=column_widths[i],
                               rotation=0,
                               aligment=Aligment.MTEXT_MIDDLE_CENTER.value)
                x_mtext += column_widths[i] / 2
            y_mtext += row_height[j] / 2
            elements["texts_row" + str(j)] = texts
            elements["all_elements"] += texts

        return elements

    def draw_table(self, document: Drawing,
                   x: float = None,
                   y: float = None,
                   row_height: float = None,
                   column_widths: list = None) -> dict:
        if x is None:
            x = self.x

        if y is None:
            y = self.y + self.columns_height

        if row_height is None:
            row_height = self.row_heights

        if column_widths is None:
            column_widths = max_per_position(self.column_widths, self.column_width_labels)

        elements = {"labels": self.draw_labels(document=document,
                                               x=x,
                                               y=y + self.columns_height,
                                               column_widths=column_widths),
                    "content": self.draw_content(document=document,
                                                 x=x,
                                                 y=y,
                                                 row_height=row_height,
                                                 column_widths=column_widths)}
        # Setting groups of elements in dictionary.
        elements["all_elements"] = elements["labels"]["all_elements"] + elements["content"]["all_elements"]

        return elements



# -*- coding: utf-8 -*-

# Local imports.
from etacad.bar import Bar
from etacad.cadtable import CADTable
from etacad.concrete import Concrete
from etacad.drawing_utils import delimit_axe, dim_linear, rect, text
from etacad.globals import (COLUMN_SET_TRANSVERSE, COLUMN_SET_LONG_REBAR, ColumnTypes, Direction, ElementTypes,
                            Orientation, CONCRETE_WEIGHT, COLUMN_SET_LONG, COLUMN_SET_TRANSVERSE_REBAR)
from etacad.stirrup import Stirrup
from etacad.utils import gen_symmetric_list, gen_position_bars

# External imports.
from attrs import define, field
from ezdxf.document import Drawing
from ezdxf.gfxattribs import GfxAttribs
from itertools import chain


@define
class Column:
    """
    Column element, computes geometrics and physics props and manages dxf drawing methods (longitudinal, transversal,
    reinforcement detailing, etc.)

    :param width: Width of the column (rectangular section) in units of length.
    :type width: float
    :param depth: Depth of the column (rectangular section) in units of length.
    :type depth: float
    :param height: Height of the column in units of length.
    :type height: float
    :param diameter: Diameter of the column if circular (optional), defaults to None.
    :type diameter: float, optional
    :param column_type: Type of column section (rectangular or circular), defaults to RECTANGULAR.
    :type column_type: int
    :param x: X-coordinate of the column's position, defaults to 0.
    :type x: float
    :param y: Y-coordinate of the column's position, defaults to 0.
    :type y: float
    :param direction: Orientation of the column (VERTICAL or HORIZONTAL), defaults to VERTICAL.
    :type direction: Direction
    :param orientation: Column's orientation (RIGHT, LEFT, etc.), defaults to RIGHT.
    :type orientation: Orientation
    :param as_sup: Dictionary representing the longitudinal steel in the upper part of the column.
    :type as_sup: dict, optional
    :param as_right: Dictionary representing the longitudinal steel in the right side of the column.
    :type as_right: dict, optional
    :param as_inf: Dictionary representing the longitudinal steel in the lower part of the column.
    :type as_inf: dict, optional
    :param as_left: Dictionary representing the longitudinal steel in the left side of the column.
    :type as_left: dict, optional
    :param anchor_sup: List of anchorage values for the upper part of the column.
    :type anchor_sup: list, optional
    :param anchor_right: List of anchorage values for the right side of the column.
    :type anchor_right: list, optional
    :param anchor_inf: List of anchorage values for the lower part of the column.
    :type anchor_inf: list, optional
    :param anchor_left: List of anchorage values for the left side of the column.
    :type anchor_left: list, optional
    :param cover: Concrete cover of the column in units of length.
    :type cover: float, optional
    :param concrete_specific_weight: Specific weight of the concrete used in the column.
    :type concrete_specific_weight: float
    :param stirrups_db: List of diameters of the stirrups used in the column.
    :type stirrups_db: list, optional
    :param stirrups_anchor: List of anchorage values for the stirrups.
    :type stirrups_anchor: list, optional
    :param stirrups_sep: List of spacing values between stirrups.
    :type stirrups_sep: list, optional
    :param stirrups_length: List of lengths of stirrups.
    :type stirrups_length: list, optional
    :param stirrups_x: List of x-coordinates for stirrups placement.
    :type stirrups_x: list, optional
    :param beams: List of crossing beams interacting with the column.
    :type beams: list, optional
    :param beams_pos: List of positions for the crossing beams.
    :type beams_pos: list, optional
    :param beams_symbol: List of symbols used to represent crossing beams.
    :type beams_symbol: list, optional
    :param nomenclature: Column identification string or nomenclature.
    :type nomenclature: str, optional

    :ivar as_sup: Dictionary containing the top longitudinal reinforcement details.
    :vartype as_sup: dict
    :ivar as_right: Dictionary containing the right longitudinal reinforcement details.
    :vartype as_right: dict
    :ivar as_inf: Dictionary containing the bottom longitudinal reinforcement details.
    :vartype as_inf: dict
    :ivar as_left: Dictionary containing the left longitudinal reinforcement details.
    :vartype as_left: dict

    :ivar max_db_sup: Maximum diameter of the top longitudinal bars.
    :vartype max_db_sup: float
    :ivar max_db_right: Maximum diameter of the right longitudinal bars.
    :vartype max_db_right: float
    :ivar max_db_inf: Maximum diameter of the bottom longitudinal bars.
    :vartype max_db_inf: float
    :ivar max_db_left: Maximum diameter of the left longitudinal bars.
    :vartype max_db_left: float
    :ivar max_db_hz: Maximum horizontal diameter among the longitudinal bars.
    :vartype max_db_hz: float
    :ivar max_db_vt: Maximum vertical diameter among the longitudinal bars.
    :vartype max_db_vt: float

    :ivar anchor_sup: List of top bar anchorage lengths.
    :vartype anchor_sup: list
    :ivar anchor_right: List of right bar anchorage lengths.
    :vartype anchor_right: list
    :ivar anchor_inf: List of bottom bar anchorage lengths.
    :vartype anchor_inf: list
    :ivar anchor_left: List of left bar anchorage lengths.
    :vartype anchor_left: list

    :ivar bars_as_sup: List of bars in the top longitudinal reinforcement.
    :vartype bars_as_sup: list
    :ivar bars_as_right: List of bars in the right longitudinal reinforcement.
    :vartype bars_as_right: list
    :ivar bars_as_inf: List of bars in the bottom longitudinal reinforcement.
    :vartype bars_as_inf: list
    :ivar bars_as_left: List of bars in the left longitudinal reinforcement.
    :vartype bars_as_left: list

    :ivar number_init_sup: Initial number for top longitudinal bars.
    :vartype number_init_sup: int
    :ivar number_init_right: Initial number for right longitudinal bars.
    :vartype number_init_right: int
    :ivar number_init_inf: Initial number for bottom longitudinal bars.
    :vartype number_init_inf: int
    :ivar number_init_left: Initial number for left longitudinal bars.
    :vartype number_init_left: int

    :ivar cover: Concrete cover thickness.
    :vartype cover: float

    :ivar stirrups_db: List of stirrup diameters.
    :vartype stirrups_db: list
    :ivar stirrups_anchor: List of stirrup anchorage lengths.
    :vartype stirrups_anchor: list
    :ivar stirrups_sep: List of stirrup separations.
    :vartype stirrups_sep: list
    :ivar stirrups_length: List of stirrup lengths.
    :vartype stirrups_length: list
    :ivar stirrups_x: List of stirrup X-coordinates.
    :vartype stirrups_x: list
    :ivar stirrups: List of stirrup elements.
    :vartype stirrups: list

    :ivar beams: List of crossing beams.
    :vartype beams: list
    :ivar beams_pos: List of crossing beams' positions.
    :vartype beams_pos: list
    :ivar beams_symbol: List of crossing beams' symbols.
    :vartype beam_symbol: list

    :ivar all_bars: List of all bars in the column.
    :vartype all_bars: list
    :ivar all_elements: List of all elements in the column.
    :vartype all_elements: list

    :ivar box_width: Width of the bounding box for the column.
    :vartype box_width: float
    :ivar box_height: Height of the bounding box for the column.
    :vartype box_height: float

    :ivar nomenclature: Nomenclature for the column reinforcement.
    :vartype nomenclature: str
    :ivar element_type: Type of element, set to COLUMN by default.
    :vartype element_type: int
    """
    # Geometric attributes.
    width: float = field(converter=float)
    depth: float = field(converter=float)
    height: float = field(converter=float)
    diameter: float = field(default=None)
    column_type: int = field(default=ColumnTypes.RECTANGULAR)
    x: float = field(default=0, converter=float)
    y: float = field(default=0, converter=float)
    direction: Direction = field(default=Direction.VERTICAL)
    orientation: Orientation = field(default=Orientation.RIGHT)

    # Longitudinal steel attributes.
    as_sup: dict = field(default=None)
    as_right: dict = field(default=None)
    as_inf: dict = field(default=None)
    as_left: dict = field(default=None)

    max_db_sup: float = field(init=False)
    max_db_right: float = field(init=False)
    max_db_inf: float = field(init=False)
    max_db_left: float = field(init=False)
    max_db_hz: float = field(init=False)
    max_db_vt: float = field(init=False)

    anchor_sup: list = field(default=0)
    anchor_right: list = field(default=0)
    anchor_inf: list = field(default=0)
    anchor_left: list = field(default=0)

    bars_as_sup: list = field(init=False)
    bars_as_right: list = field(init=False)
    bars_as_inf: list = field(init=False)
    bars_as_left: list = field(init=False)

    number_init_sup: int = field(init=False)
    number_init_right: int = field(init=False)
    number_init_inf: int = field(init=False)
    number_init_left: int = field(init=False)

    cover: float = field(default=0, converter=float)

    # Concrete attributes.
    concrete: Concrete = field(init=False)
    concrete_specific_weight: float = CONCRETE_WEIGHT

    # Stirrup attributes.
    stirrups_db: list = field(default=None)
    stirrups_anchor: list = field(default=None)
    stirrups_sep: list = field(default=None)
    stirrups_length: list = field(default=None)
    stirrups_x: list = field(default=None)
    stirrups: list = field(init=False)

    # Crossing beams attributes.
    beams: list = field(default=None)
    beams_pos: list = field(default=None)
    beams_symbol: list = field(default=None)

    # Entities groups.
    all_bars: list = field(init=False)
    all_elements: list = field(init=False)

    # Box attributes.
    box_width: float = field(init=False)
    box_height: float = field(init=False)

    # Position bar attributes.
    nomenclature: str = field(default="#")
    positions: dict = field(init=False)

    # Column attributes.
    denomination: str = field(default=None)
    number_init: int = field(default=None)
    element_type: ElementTypes = field(default=ElementTypes.COLUMN)

    def __attrs_post_init__(self):
        # Longitudinal steel attributes.
        if self.number_init is None:
            self.number_init = 0

        # Top bars.
        if self.as_sup:
            self.max_db_sup = max(self.as_sup.keys())
            self.number_init_sup = self.number_init if self.number_init else 0
            self.number_init = self.number_init + len(self.as_sup) if self.number_init else len(self.as_sup)

            self.anchor_sup = self.anchor_sup if type(self.anchor_sup) == list else [self.anchor_sup] * sum(
                self.as_sup.values())
        else:
            self.as_sup, self.max_db_sup, self.number_init_sup = {}, 0, 0

        # Right bars.
        if self.as_right:
            self.max_db_right = max(self.as_right.keys())
            self.number_init_right = self.number_init if self.number_init is not None else 0
            self.number_init = self.number_init + len(self.as_right) if self.number_init else len(self.as_right)

            self.anchor_right = self.anchor_right if type(self.anchor_right) == list else [self.anchor_right] * sum(
                self.as_right.values())
        else:
            self.as_right, self.max_db_right, self.number_init_right = {}, 0, 0

        # Inferior bars.
        if self.as_inf:
            self.max_db_inf = max(self.as_inf.keys())
            self.number_init_inf = self.number_init if self.number_init is not None else 0
            self.number_init = self.number_init + len(self.as_inf) if self.number_init else len(self.as_inf)

            self.anchor_inf = self.anchor_inf if type(self.anchor_inf) == list else [self.anchor_inf] * sum(
                self.as_inf.values())
        else:
            self.as_inf, self.max_db_inf, self.number_init_inf = {}, 0, 0

        # Left bars.
        if self.as_left:
            self.max_db_left = max(self.as_left.keys())
            self.number_init_left = self.number_init if self.number_init is not None else 0
            self.number_init = self.number_init + len(self.as_left) if self.number_init else len(self.as_left)

            self.anchor_left = self.anchor_left if type(self.anchor_left) == list else [self.anchor_left] * sum(
                self.as_left.values())
        else:
            self.as_left, self.max_db_left, self.number_init_left = {}, 0, 0

        self.max_db_hz = max(self.max_db_sup, self.max_db_inf)
        self.max_db_vt = max(self.max_db_right, self.max_db_left)

        self.bars_as_sup = self.__dict_to_bars(bars=self.as_sup,
                                               width=self.width,
                                               x=self.x,
                                               y=self.y,
                                               side=Orientation.TOP,
                                               anchor=self.anchor_sup,
                                               nomenclature=self.nomenclature,
                                               number_init=self.number_init_sup) if self.as_sup else []
        self.bars_as_right = self.__dict_to_bars(bars=self.as_right,
                                                 width=self.depth,
                                                 x=self.x,
                                                 y=self.y,
                                                 side=Orientation.RIGHT,
                                                 anchor=self.anchor_right,
                                                 nomenclature=self.nomenclature,
                                                 number_init=self.number_init_right) if self.as_sup else []
        self.bars_as_inf = self.__dict_to_bars(bars=self.as_inf,
                                               width=self.width,
                                               x=self.x,
                                               y=self.y,
                                               side=Orientation.BOTTOM,
                                               anchor=self.anchor_inf,
                                               nomenclature=self.nomenclature,
                                               number_init=self.number_init_inf) if self.as_sup else []
        self.bars_as_left = self.__dict_to_bars(bars=self.as_left,
                                                width=self.depth,
                                                x=self.x,
                                                y=self.y,
                                                side=Orientation.LEFT,
                                                anchor=self.anchor_left,
                                                nomenclature=self.nomenclature,
                                                number_init=self.number_init_left) if self.as_sup else []

        # Concrete attributes.
        vertices = [(0, 0),
                    (0, self.depth),
                    (self.width, self.depth),
                    (self.width, 0)]

        self.concrete = Concrete(vertices=vertices,
                                 height=self.height,
                                 x=self.x,
                                 y=self.y,
                                 specific_weight=self.concrete_specific_weight)

        # Stirrups attributes.
        if self.stirrups_db is not None:
            if type(self.stirrups_db) == float:
                self.stirrups_db = [self.stirrups_db]

            if self.stirrups_anchor is None:
                self.stirrups_anchor = [0.1] * len(self.stirrups_db)

            elif type(self.stirrups_anchor) == float:
                self.stirrups_anchor = [self.stirrups_anchor] * len(self.stirrups_db)

            if type(self.stirrups_sep) == float:
                self.stirrups_sep = [self.stirrups_sep] * len(self.stirrups_db)

            if self.stirrups_length is None:
                self.stirrups_length = [self.height - (self.depth + self.cover) * 2]

            if self.stirrups_x is None:
                self.stirrups_x = [self.depth + self.cover]

        self.stirrups = self.__list_to_stirrups()

        # Position bar attributes.
        min_number_init = min(self.number_init_sup, self.number_init_right, self.number_init_inf, self.number_init_left)

        self.positions = gen_position_bars(dictionaries=[self.as_sup, self.as_right, self.as_inf, self.as_left],
                                           nomenclature=self.nomenclature,
                                           number_init=min_number_init)

        # Entities groups.
        self.all_bars = self.bars_as_sup + self.bars_as_right + self.bars_as_inf + self.bars_as_left
        self.all_elements = self.all_bars + self.stirrups

        # Box attributes.
        self.box_width = self.width
        self.box_height = self.height

    def draw_longitudinal(self, document: Drawing,
                          x: float = None,
                          y: float = None,
                          concrete_shape: bool = True,
                          bars: bool = True,
                          beams: bool = True,
                          beams_axes: bool = True,
                          stirrups: bool = True,
                          middle_axe: bool = True,
                          middle_axe_symbol: str = "A",
                          dim: bool = True,
                          dim_style: str = "EZ_M_25_H25_CM",
                          unifilar_bars: bool = False,
                          unifilar_stirrups: bool = True) -> dict:
        """
        Draws the longitudinal view of the column, including concrete shape, beams,
        stirrups, and bars. Also includes dimensioning and optional middle axes.

        :param document: The drawing document to which the elements will be added.
        :type document: Drawing
        :param x: X-coordinate for the drawing position. Defaults to column's x position.
        :type x: float, optional
        :param y: Y-coordinate for the drawing position. Defaults to column's y position.
        :type y: float, optional
        :param concrete_shape: Whether to draw the concrete shape. Defaults to True.
        :type concrete_shape: bool
        :param bars: Whether to draw the longitudinal bars. Defaults to True.
        :type bars: bool
        :param beams: Whether to draw crossing beams. Defaults to True.
        :type beams: bool
        :param beams_axes: Whether to draw the axes for beams. Defaults to True.
        :type beams_axes: bool
        :param stirrups: Whether to draw the stirrups. Defaults to True.
        :type stirrups: bool
        :param middle_axe: Whether to draw the middle axis. Defaults to True.
        :type middle_axe: bool
        :param middle_axe_symbol: Symbol to use for the middle axis. Defaults to "A".
        :type middle_axe_symbol: str
        :param dim: Whether to add dimensions to the drawing. Defaults to True.
        :type dim: bool
        :param dim_style: Style of the dimensions. Defaults to "EZ_M_25_H25_CM".
        :type dim_style: str
        :param unifilar_bars: Whether to draw bars in unifilar view. Defaults to False.
        :type unifilar_bars: bool
        :param unifilar_stirrups: Whether to draw stirrups in unifilar view. Defaults to True.
        :type unifilar_stirrups: bool
        :return: A dict of entities drawn on the document.
        :rtype: dict
        """
        if x is None:
            x = self.x
        if y is None:
            y = self.y

        elements = {"concrete": [],
                    "beam_elements": [],
                    "beam_axe_elements": [],
                    "column_axe_elements": [],
                    "dimensions_elements": [],
                    "stirrups": [],
                    "bars": []}

        delimit_axe_width = self.width * 4
        delimit_axe_x = x - self.width * 2

        # Drawing concrete shape.
        if concrete_shape:
            elements["concrete"] = self.concrete.draw_longitudinal(document=document,
                                                                   x=x,
                                                                   y=y,
                                                                   dimensions=dim,
                                                                   dimensions_inner=False,
                                                                   settings=COLUMN_SET_LONG["concrete_settings"])

        # Drawing beams.
        if self.beams:
            if beams:
                for i, y_beam in enumerate(self.beams_pos):
                    y_column_relative = y + y_beam
                    y_delimit_axe_relative = y_column_relative + self.beams[i][1] / 2
                    if beams:
                        elements["beam_elements"] += rect(doc=document,
                                                          width=self.beams[i][0],
                                                          height=self.beams[i][1],
                                                          x=x,
                                                          y=y_column_relative,
                                                          fill=True,
                                                          polyline=True)
                    if beams_axes:
                        elements["beam_axe_elements"] += delimit_axe(document=document,
                                                                     x=delimit_axe_x,
                                                                     y=y_delimit_axe_relative,
                                                                     height=delimit_axe_width,
                                                                     radius=0.1,
                                                                     text_height=0.1,
                                                                     symbol=self.beams_symbol[i],
                                                                     direction=Direction.HORIZONTAL,
                                                                     attr=GfxAttribs(linetype="CENTER"))

        # Drawing axe.
        if middle_axe:
            elements["column_axe_elements"] += delimit_axe(document=document,
                                                           x=x + self.width / 2,
                                                           y=y - self.width * 2,
                                                           height=self.height + self.width * 4,
                                                           symbol=middle_axe_symbol,
                                                           direction=Direction.VERTICAL,
                                                           attr=GfxAttribs(linetype="CENTER"))

        # Drawing of stirrups.
        if stirrups:
            for stirrup in self.stirrups:
                elements["stirrups"].append(stirrup.draw_longitudinal(document=document,
                                                                      x=x + (stirrup.x - self.x),
                                                                      y=y + (stirrup.y - self.y),
                                                                      unifilar=unifilar_stirrups))
            # Drawing dimensions.
            if dim:
                for stirrup in self.stirrups:
                    elements["dimensions_elements"] += dim_linear(document=document,
                                                                  p_base=(x - self.width * 3,
                                                                          y + (stirrup.y - self.y) + stirrup.reinforcement_length / 2),
                                                                  p1=(x, y + (stirrup.y - self.y)),
                                                                  p2=(x, y + (stirrup.y - self.y) + stirrup.reinforcement_length),
                                                                  rotation=90,
                                                                  dimstyle=dim_style)

        # Drawing of bars.
        if self.as_sup or self.as_right or self.as_inf or self.as_left:
            if bars:
                for bar in self.bars_as_inf:
                    delta_unfilar = 0 if not unifilar_bars else - bar.radius
                    elements["bars"].append(bar.draw_longitudinal(document=document,
                                                                  x=x + (bar.x - self.x) + delta_unfilar,
                                                                  y=y + (bar.y - self.y),
                                                                  unifilar=unifilar_bars,
                                                                  dimensions=False,
                                                                  denomination=False))  # Only left bars.

        # Setting groups of elements in dictionary.
        elements["all_elements"] = (elements["concrete"]["all_elements"] +
                                    elements["beam_elements"] +
                                    elements["beam_axe_elements"] +
                                    elements["column_axe_elements"] +
                                    elements["dimensions_elements"] +
                                    list(chain(*[bar_dict["all_elements"] for bar_dict in elements["bars"]])) +
                                    list(chain(*[st_dict["all_elements"] for st_dict in elements["stirrups"]])))

        return elements

    def draw_transverse(self,
                        document: Drawing,
                        x: float = None,
                        y: float = None,
                        y_section: float = None,
                        unifilar: bool = False,
                        dimensions: bool = True,
                        settings: dict = COLUMN_SET_TRANSVERSE) -> dict:
        """
        Draws the transverse view of the column at a given y-section.
        generate a drawing from the transverse perspective.

        :param document: The DXF document where the column will be drawn.
        :type document: Drawing
        :param x: X-coordinate of the starting point for drawing.
        :type x: float
        :param y: Y-coordinate of the starting point for drawing.
        :type y: float
        :param y_section: The y-coordinate of the section to be drawn.
        :type y_section: float, optional
        :param unifilar: If True, the bars are drawn as unifilar.
        :type unifilar: bool
        :param dimensions: If True, dimensions are drawn.
        :type dimensions: bool
        :param settings: Dict with column transverse drawing settings.
        :type settings: dict
        :return: A dict of entities representing the transverse view of the column.
        :rtype: dict
        """
        if x is None:
            x = self.x
        if y is None:
            y = self.y

        if y_section is None:
            y_section = self.height / 2

        # Checking if y given is in column height.
        if not 0 <= y_section <= self.height:
            return []

        elements = {"concrete": None,
                    "bars": [],
                    "stirrups": []}

        # Obtaining elements for given y_section.
        entities = self.__elements_section(y=y_section)

        # Filtering elements.
        bars = filter(lambda e: e.element_type == ElementTypes.BAR, entities)
        stirrups = filter(lambda e: e.element_type == ElementTypes.STIRRUP, entities)

        # Concrete shape drawing.
        elements["concrete"] = self.concrete.draw_transverse(document=document,
                                                             x=x,
                                                             y=y,
                                                             dimensions=dimensions,
                                                             dimensions_boxing=True,
                                                             dimensions_inner=False,
                                                             settings=settings["concrete_settings"])

        # Drawing of bars.
        for bar in bars:
            elements["bars"].append(bar.draw_transverse(document=document, x=x, y=y))

        # Drawing of stirrups.
        for stirrup in stirrups:
            delta_x = self.cover - max(self.max_db_sup, self.max_db_inf) / 2 - stirrup.diameter
            delta_y = self.cover - self.max_db_inf / 2 - stirrup.diameter

            elements["stirrups"].append(stirrup.draw_transverse(document=document,
                                                                x=x + delta_x,
                                                                y=y + delta_y,
                                                                unifilar=unifilar))

        # Setting groups of elements in dictionary.
        elements["all_elements"] = (elements["concrete"]["all_elements"] +
                                    list(chain(*[bar_dict["all_elements"] for bar_dict in elements["bars"]])) +
                                    list(chain(*[st_dict["all_elements"] for st_dict in elements["stirrups"]])))

        return elements

    def draw_longitudinal_rebar_detailing(self,
                                          document: Drawing,
                                          x: float = None,
                                          y: float = None,
                                          unifilar: bool = True,
                                          beam_axes: bool = True,
                                          settings: dict = COLUMN_SET_LONG_REBAR) -> dict:
        """
        Draws the longitudinal rebar detailing for the column.

        :param document: The DXF document where the detailing will be drawn.
        :type document: Drawing
        :param x: X-coordinate of the starting point for drawing.
        :type x: float, optional
        :param y: Y-coordinate of the starting point for drawing.
        :type y: float, optional
        :param unifilar: If True, the rebar is drawn as unifilar.
        :type unifilar: bool
        :param beam_axes: If True, the axes of the beam are drawn.
        :type beam_axes: bool
        :param settings: Dict with column longitudinal rebar drawing settings.
        :type settings: dict
        :return: A dict of graphical entities representing the longitudinal rebar detailing.
        :rtype: dict
        """

        elements = {"text_elements": [],
                    "bars": [],
                    "barline_elements": [],
                    "beam_axes_elements": [],
                    "all_elements": []}

        def __draw_section_bars(bars: list = None,
                                spacing: float = settings["spacing"],
                                rebar_x: float = None,
                                rebar_y: float = None,
                                barline=True,
                                text_reference: str = None) -> tuple:

            bars_elements = []
            text_elements = []
            barline_elements = []

            if text_reference:
                text_elements += text(document=document,
                                      text=text_reference,
                                      height=settings["text_height"],
                                      point=(rebar_x + 0.02, rebar_y + self.height + spacing * 2))

            unique_bars_list = []
            for bar in bars:
                if bar.denomination not in unique_bars_list:
                    rebar_x += spacing + bar.box_height
                    bars_elements.append(bar.draw_longitudinal(document=document,
                                                               x=rebar_x,
                                                               y=rebar_y,
                                                               unifilar=unifilar,
                                                               dimensions=True,
                                                               settings=settings["bar_settings"]))
                    unique_bars_list.append(bar.denomination)

            if barline:
                barline_elements += rect(doc=document,
                                         width=0,
                                         height=self.height + self.depth * 4,
                                         x=rebar_x + spacing,
                                         y=rebar_y - spacing,
                                         sides=(0, 1, 0, 0))

            rebar_x += spacing

            return text_elements, bars_elements, barline_elements, rebar_x, rebar_y

        group, rebar_x, rebar_y = [], x, y
        if self.as_sup:
            barline = True if self.as_right or self.bars_as_left or self.bars_as_inf else False
            elements_tuple = __draw_section_bars(bars=self.bars_as_sup,
                                                 spacing=settings["spacing"],
                                                 rebar_x=rebar_x,
                                                 rebar_y=rebar_y,
                                                 barline=barline,
                                                 text_reference="As sup.")

            text_elements, bars, barline_elements, rebar_x, rebar_y = elements_tuple

            elements["text_elements"].extend(text_elements)
            elements["bars"].extend(bars)
            elements["barline_elements"].extend(barline_elements)

        if self.as_right:
            barline = True if self.bars_as_left or self.bars_as_inf else False
            elements_tuple = __draw_section_bars(bars=self.bars_as_right,
                                                 spacing=settings["spacing"],
                                                 rebar_x=rebar_x,
                                                 rebar_y=rebar_y,
                                                 barline=barline,
                                                 text_reference="As right.")

            text_elements, bars, barline_elements, rebar_x, rebar_y = elements_tuple

            elements["text_elements"].extend(text_elements)
            elements["bars"].extend(bars)
            elements["barline_elements"].extend(barline_elements)

        if self.as_left:
            barline = True if self.bars_as_inf else False
            elements_tuple = __draw_section_bars(bars=self.bars_as_left,
                                                 spacing=settings["spacing"],
                                                 rebar_x=rebar_x,
                                                 rebar_y=rebar_y,
                                                 barline=barline,
                                                 text_reference="As left.")

            text_elements, bars, barline_elements, rebar_x, rebar_y = elements_tuple

            elements["text_elements"].extend(text_elements)
            elements["bars"].extend(bars)
            elements["barline_elements"].extend(barline_elements)

        if self.as_inf:
            elements_tuple = __draw_section_bars(bars=self.bars_as_inf,
                                                 spacing=settings["spacing"],
                                                 rebar_x=rebar_x,
                                                 rebar_y=rebar_y,
                                                 barline=False,
                                                 text_reference="As inf.")

            text_elements, bars, barline_elements, rebar_x, rebar_y = elements_tuple

            elements["text_elements"].extend(text_elements)
            elements["bars"].extend(bars)
            elements["barline_elements"].extend(barline_elements)

        # if beam_axes:
        #     for i, x_column in enumerate(self.beams_pos):
        #         delimit_axe_height = y - rebar_y
        #         delimit_axe_y = rebar_y
        #         x_column_relative = x + x_column
        #         x_delimit_axe_relative = x_column_relative + self.columns[i][0] / 2
        #
        #         group += delimit_axe(document=document,
        #                              x=x_delimit_axe_relative,
        #                              y=delimit_axe_y,
        #                              height=delimit_axe_height,
        #                              radius=0.1,
        #                              text_height=0.1,
        #                              attr=GfxAttribs(linetype="CENTER"))

        # Setting groups of elements in dictionary.
        elements["all_elements"] = (elements["text_elements"] +
                                    list(chain(*[bar_dict["all_elements"] for bar_dict in elements["bars"]])) +
                                    elements["barline_elements"] +
                                    elements["beam_axes_elements"])

        return elements

    def draw_transverse_rebar_detailing(self, document: Drawing,
                                        x: float = None,
                                        y: float = None,
                                        y_section: float = None,
                                        unifilar: bool = False,
                                        dimensions: bool = True,
                                        settings: dict = COLUMN_SET_TRANSVERSE_REBAR) -> list:
        """
        Draws the transverse rebar detailing for the column at a given y-section.

        :param document: The DXF document where the detailing will be drawn.
        :type document: Drawing
        :param x: X-coordinate of the starting point for drawing.
        :type x: float, optional
        :param y: Y-coordinate of the starting point for drawing.
        :type y: float, optional
        :param y_section: The y-coordinate of the section to be drawn.
        :type y_section: float, optional
        :param unifilar: If True, the rebar is drawn as unifilar.
        :type unifilar: bool
        :param dimensions: If True, dimensions are drawn.
        :type dimensions: bool
        :param settings: Dict with column longitudinal rebar drawing settings.
        :type settings: dict
        :return: A dict of graphical entities representing the transverse rebar detailing.
        :rtype: dict
        """
        if y_section is None:
            y_section = self.height / 2

        # Checking if x given is in beam length.
        if not 0 <= y_section <= self.height:
            return []

        elements = {"stirrups": []}

        stirrups = self.__elements_section(elements=self.stirrups,
                                           y=y_section)

        group = []
        for stirrup in stirrups:
            elements["stirrups"].append(stirrup.draw_transverse(document=document,
                                                                x=x,
                                                                y=y,
                                                                unifilar=unifilar,
                                                                dimensions=dimensions))

        # Setting groups of elements in dictionary.
        elements["all_elements"] = (list(chain(*[st_dict["all_elements"] for st_dict in elements["stirrups"]])))

        return elements

    def __dict_to_bars(self, bars: dict,
                       width: float,
                       x: float,
                       y: float,
                       side: Orientation,
                       anchor: list,
                       nomenclature: str = None,
                       number_init: int = None) -> list:
        """
        Converts a dictionary of bars into a list of Bar objects, placing them according to the specified side of
        the column.

        :param bars: Dictionary containing bar information.
        :type bars: dict
        :param width: Width of the column section.
        :type width: float
        :param x: X-coordinate of the starting point for placing the bars.
        :type x: float
        :param y: Y-coordinate of the starting point for placing the bars.
        :type y: float
        :param side: Indicates the side of the column where the bars are placed (0: top, 1: right, 2: bottom, 3: left).
        :type side: int
        :param anchor: List of anchor lengths for each bar.
        :type anchor: list
        :param nomenclature: Nomenclature for the bars.
        :type nomenclature: str, optional
        :param number_init: Initial number for bar numbering.
        :type number_init: int, optional
        :return: A list of Bar objects representing the bars placed in the specified section of the column.
        :rtype: list
        :raises ValueError: If the side is not within the range [0, 3].
        """
        # Creating symetric list of bars.
        list_bars, list_denom, list_pos, list_quantity = gen_symmetric_list(dictionary=bars,
                                                                            nomenclature=nomenclature,
                                                                            number_init=number_init)

        # Calculating separation and definition of initial x_sep and y_sep.
        if side == Orientation.TOP or side == Orientation.BOTTOM:  # Sides top (0) or bottom (2).
            db_max = max(self.max_db_sup, self.max_db_inf)
            width_util = width - self.cover * 2 + (db_max - max(list_bars))
            separation = width_util / (len(list_bars) - 1)
            x_sep, y_sep = -separation, 0

        elif side == Orientation.RIGHT or side == Orientation.LEFT:  # Sides right (1) or left (3).
            db_max = max([self.max_db_right, self.max_db_left])
            width_util = width - self.cover * 2
            separation = width_util / (len(list_bars) + 1)
            x_sep, y_sep = 0, 0
        else:
            raise ValueError

        # Generating bars.
        entities, delta_x, delta_y, orientation = [], 0, 0, Orientation.BOTTOM
        for i, db in enumerate(list_bars):
            # Calculation of x_delta and y_delta.
            if side == Orientation.TOP or side == Orientation.BOTTOM:
                x_sep += separation
                delta_x = self.cover - db_max / 2 + (db_max / 2 - db / 2)
                if side == Orientation.TOP:
                    delta_y_transverse = self.depth - self.cover + self.max_db_sup / 2 - db
                if side == Orientation.BOTTOM:
                    delta_y_transverse = self.cover - self.max_db_inf / 2
            elif side == Orientation.RIGHT or side == Orientation.LEFT:
                y_sep += separation
                delta_y = self.cover - db / 2
                if side == Orientation.RIGHT:
                    delta_x = self.width - (self.cover + db - max( [self.max_db_sup, self.max_db_inf, self.max_db_right]) / 2)
                    delta_y_transverse = delta_y
                if side == Orientation.LEFT:
                    delta_x = self.cover - max([self.max_db_sup, self.max_db_inf, self.max_db_left]) / 2
                    delta_y_transverse = delta_y
            else:
                raise ValueError

            x_bar_long = x + x_sep + delta_x
            x_bar_transverse = x_sep + delta_x
            y_bar_long = y + self.cover
            y_bar_transverse = y_sep + delta_y_transverse

            entities.append(Bar(reinforcement_length=self.height - self.cover * 2,
                                diameter=db,
                                x=x_bar_long,
                                y=y_bar_long,
                                left_anchor=anchor[i],
                                right_anchor=anchor[i],
                                mandrel_radius=db,
                                direction=Direction.VERTICAL,
                                orientation=Orientation.BOTTOM,
                                transverse_center=(x_bar_transverse, y_bar_transverse),
                                denomination=list_denom[i],
                                position=list_pos[i],
                                quantity=list_quantity[i]))

        return entities

    def __elements_section(self,
                           elements: list = None,
                           y: float = None) -> list:
        """
        Filter elements that are in the Y coordinate given.

        :param elements: Group of entities (bars and stirrups).
        :type elements: list
        :param y: Y coordinate of the section.
        :type y: float
        :return: Group of entities that are in the Y coordinate given.
        :rtype: list
        """
        if elements is None:
            elements = self.all_elements

        if y is None:
            y = self.y + self.height / 2  # Default value if is nothing entered.

        # Filtering elements by X coordinate.
        entities = []
        for element in elements:
            if element.y <= self.y + y <= element.y + element.reinforcement_length:
                entities.append(element)

        return entities

    def __list_to_stirrups(self) -> list:
        """
        Generates Stirrup instances of data given at initialization.

        :return: List of Stirrup instances.
        :rtype: list
        """
        if self.stirrups_db is None:
            return []

        stirrup_width = self.width - self.cover * 2 + max(self.max_db_sup, self.max_db_inf)
        stirrup_height = self.depth - (self.cover * 2 - self.max_db_sup / 2 - self.max_db_inf / 2)
        stirrups = zip(self.stirrups_db, self.stirrups_length, self.stirrups_sep, self.stirrups_anchor, self.stirrups_x)
        mandrel_radius_sup = self.max_db_sup / 2
        mandrel_radius_inf = self.max_db_inf / 2

        # Generating stirrups.
        entities = []  # Empty list that will contain stirrups elements.
        n = 0
        for stirrup in stirrups:
            stirrup_x = self.x + self.cover - self.max_db_inf / 2 - stirrup[0]  # Subract of thick of stirrup.
            stirrup_y = self.y + stirrup[4]  # Difining y coordinate.

            # Creating of stirrup.

            entities.append(Stirrup(width=stirrup_width + stirrup[0] * 2,
                                    height=stirrup_height + stirrup[0] * 2,
                                    diameter=stirrup[0],
                                    reinforcement_length=stirrup[1],
                                    spacing=stirrup[2],
                                    x=stirrup_x,
                                    y=stirrup_y,
                                    mandrel_radius_top=mandrel_radius_sup,
                                    mandrel_radius_bottom=mandrel_radius_inf,
                                    anchor=stirrup[3],
                                    direction=Direction.VERTICAL,
                                    position="{0}S{1}".format(self.nomenclature, n)))
            n += 1

        return entities

    def draw_table_rebar_detailing(self,
                                   document: Drawing,
                                   x: float = None,
                                   y: float = None) -> dict:
        # Getting data.
        data = self.extract_data()

        # Creating table.
        table = CADTable(data=data, labels=["POSITION", "DIAMETER", "SPACING", "QUANTITY", "LENGTH", "TOTAL LENGTH",
                                            "WEIGHT"])

        # Drawing table.
        elements = table.draw_table(document=document, x=x, y=y)

        return elements

    def extract_data(self) -> list:
        data = []
        position_list = []
        for element in (self.all_bars + self.stirrups):
            if element.position not in position_list:
                position = element.position
                diameter = element.diameter
                spacing = "{0}".format(element.spacing) if element.element_type == ElementTypes.STIRRUP else "-"
                quantity = element.quantity if element.element_type == ElementTypes.STIRRUP else self.positions[element.position]["quantity"]
                length = "{0:.2f}".format(float(element.length))
                total_length = "{0:.2f}".format(element.quantity * element.length)

                weight = "{0:.2f}".format(element.quantity * element.weight)
                total_weight = 0

                data.append([position, diameter, spacing, quantity, length, total_length, weight])
                position_list.append(element.position)

        return data



# -*- coding: utf-8 -*-

# Imports.
# Locals imports.
from etacad.drawing_utils import line, polyline, translate, dim_linear
from etacad.globals import CONCRETE_WEIGHT, DRotation, CONCRETE_SET_LONG, CONCRETE_SET_TRANSVERSE, ElementTypes
from etacad.geometry.polygon import Polygon
from etacad.geometry.utils import displace_perpendicular, get_angle

# External imports.
from attrs import define, field
from ezdxf.document import Drawing


@define
class Concrete:
    """
    A class used to represent a concrete section with various geometric and physical attributes.

    :param vertices: List of tuples representing the vertices of the concrete polygon.
    :type vertices: list[tuple[float, float]]
    :param height: Height of the concrete section. Defaults to None.
    :type height: float, optional
    :param length: Length of the concrete section. Defaults to None.
    :type length: float, optional
    :param x: X-coordinate for the concrete section's placement. Defaults to 0.0.
    :type x: float, optional
    :param y: Y-coordinate for the concrete section's placement. Defaults to 0.0.
    :type y: float, optional
    :param specific_weight: Specific weight of the concrete. Defaults to CONCRETE_WEIGHT.
    :type specific_weight: float, optional
    :param element_type: Type of the structural element (e.g., CONCRETE).
    :type element_type: ElementTypes, optional

    :ivar dim3D: The third dimension (either height or length) of the concrete section.
    :vartype dim3D: float
    :ivar volume: Volume of the concrete section.
    :vartype volume: float
    :ivar weight: Weight of the concrete section.
    :vartype weight: float
    :ivar box_width: Width of the bounding box in the longitudinal direction.
    :vartype box_width: float
    :ivar box_height: Height of the bounding box in the longitudinal direction.
    :vartype box_height: float
    :ivar box_width_transverse: Width of the bounding box in the transverse direction.
    :vartype box_width_transverse: float
    :ivar box_height_transverse: Height of the bounding box in the transverse direction.
    :vartype box_height_transverse: float
    :ivar element_type: Type of the structural element (e.g., CONCRETE).
    :vartype element_type: ElementTypes, optional

    :raises TypeError: If neither height nor length is provided.

    Methods
    -------
    draw_longitudinal(document, x=None, y=None, dimensions=True, dimensions_inner=True, settings=CONCRETE_SET_LONG)
        Draws the concrete section in the longitudinal direction with optional dimensioning.

    draw_transverse(document, x=None, y=None, dimensions=True, dimensions_boxing=True, dimensions_inner=True,
     settings=CONCRETE_SET_TRANSVERSE)
        Draws the concrete section in the transverse direction with optional dimensioning.

    polygon
        Property method to return a Polygon object created from the vertices of the concrete section.
    """
    # Geometric attributes.
    vertices: list[tuple[float, float]]
    height: float = field(default=None)
    length: float = field(default=None)
    x: float = field(default=0.0)
    y: float = field(default=0.0)
    dim3D: float = field(init=False)
    # polygon: Polygon = field(init=False)

    # Physics attributes.
    volume: float = field(init=False)
    weight: float = field(init=False)
    specific_weight: float = field(default=CONCRETE_WEIGHT)

    # Boxing attributes.
    box_width: float = field(init=False)
    box_height: float = field(init=False)
    box_width_transverse: float = field(init=False)
    box_height_transverse: float = field(init=False)

    # Others.
    element_type: ElementTypes = field(default=ElementTypes.CONCRETE)

    def __attrs_post_init__(self):
        # Geometric attributes.
        if self.height is not None:
            self.dim3D = self.height
            self.box_width = self.polygon.get_right_point()[0][0] - self.polygon.get_left_point()[0][0]
            self.box_height = self.height

        elif self.length is not None:
            self.dim3D = self.length
            self.box_width = self.length
            self.box_height = self.polygon.get_top_point()[0][1] - self.polygon.get_bottom_point()[0][1]

        else:
            raise TypeError("Missing length or height argument.")

        # Physics attributes.
        self.volume = self.polygon.area * self.dim3D
        self.weight = self.volume * self.specific_weight

        self.box_width_transverse = self.polygon.get_right_point()[0][0] - self.polygon.get_left_point()[0][0]
        self.box_height_transverse = self.polygon.get_top_point()[0][1] - self.polygon.get_bottom_point()[0][1]

    @property
    def polygon(self):
        """
        Property to get a Polygon object created from the vertices of the concrete section.

        :return: A Polygon object representing the concrete section.
        :rtype: Polygon
        """
        return Polygon(vertices=self.vertices)

    def draw_longitudinal(self,
                          document: Drawing,
                          x: float = None,
                          y: float = None,
                          dimensions: bool = True,
                          dimensions_inner: bool = True,
                          settings: dict = CONCRETE_SET_LONG) -> dict:
        """
        Draws the concrete section in the longitudinal direction with optional dimensioning.

        :param document: The drawing document to which the concrete section will be added.
        :type document: Drawing
        :param x: X-coordinate for the placement of the concrete section. Defaults to None, in which case `self.x` is
         used.
        :type x: float, optional
        :param y: Y-coordinate for the placement of the concrete section. Defaults to None, in which case `self.y` is
         used.
        :type y: float, optional
        :param dimensions: Flag to indicate whether to draw dimensions. Defaults to True.
        :type dimensions: bool, optional
        :param dimensions_inner: Flag to indicate whether to draw inner dimensions. Defaults to True.
        :type dimensions_inner: bool, optional
        :param settings: Dictionary of settings for dimensioning. Defaults to `CONCRETE_SET_LONG`.
        :type settings: dict, optional

        :return: A dictionary with keys "concrete_elements", "dimensions", and "all_elements", each containing the
        corresponding drawing elements.
        :rtype: dict
        """
        if x is None:
            x = self.x
        if y is None:
            y = self.y

        elements = {}
        concrete_elements = []
        dim_elements = []

        if self.length is not None:
            # Getting right points between top and bottom points of concrete section.
            top_point = self.polygon.get_top_point()[-1]
            bottom_point = self.polygon.get_bottom_point()[-1]
            top_index = self.vertices.index(top_point)
            bottom_index = self.vertices.index(bottom_point)

            right_points = self.polygon.get_points_between_index_limits_by_dr(index1=top_index, index2=bottom_index)

            # Drawing concrete section.
            for point in right_points:
                concrete_elements += line(doc=document,
                                          p1=(x, y + point[1]),
                                          p2=(x + self.length, y + point[1]))
            concrete_elements += line(doc=document,
                                      p1=(x, y + top_point[1]),
                                      p2=(x, y + bottom_point[1]))
            concrete_elements += line(doc=document,
                                      p1=(x + self.length, y + top_point[1]),
                                      p2=(x + self.length, y + bottom_point[1]))

        if self.height is not None:
            # Getting front points between right and left points of concrete section.
            left_point = self.polygon.get_left_point()[0]
            right_point = self.polygon.get_right_point()[0]
            left_index = self.vertices.index(left_point)
            right_index = self.vertices.index(right_point)

            front_points = self.polygon.get_points_between_index_limits_by_dr(index1=right_index, index2=left_index)

            # Drawing concrete section.
            for point in front_points:
                concrete_elements += line(doc=document,
                                          p1=(x + point[0], y),
                                          p2=(x + point[0], y + self.height))
            concrete_elements += line(doc=document,
                                      p1=(x + left_point[0], y),
                                      p2=(x + right_point[0], y))
            concrete_elements += line(doc=document,
                                      p1=(x + left_point[0], y + self.height),
                                      p2=(x + right_point[0], y + self.height))

        # Drawing of dimensions.
        if dimensions:
            dim_elements = []
            if self.length:
                # Boxing dimensions.
                dim_elements += dim_linear(document=document,
                                           p_base=(x + self.box_width / 2,
                                                   y + self.box_height + settings["text_dim_distance_horizontal"]),
                                           p1=(x, y + self.box_height),
                                           p2=(x + self.box_width, y + self.box_height),
                                           dimstyle=settings["dim_style_boxing"])
                dim_elements += dim_linear(document=document,
                                           p_base=(x - settings["text_dim_distance_vertical"],
                                                   y + self.box_height / 2),
                                           p1=(x, y),
                                           p2=(x, y + self.box_height),
                                           rotation=90,
                                           dimstyle=settings["dim_style_boxing"])

                if dimensions_inner:
                    # Polygon dimensions.
                    for i in range(len(right_points) - 1):
                        dim_elements += dim_linear(document=document,
                                                   p_base=(x - settings["text_dim_distance_vertical"],
                                                           y + abs(right_points[i][1] - right_points[i + 1][1]) / 2),
                                                   p1=(x, y + right_points[i][1]),
                                                   p2=(x, y + right_points[i + 1][1]),
                                                   rotation=90,
                                                   dimstyle=settings["dim_style_inner"])

            if self.height:
                # Boxing dimensions.
                dim_elements += dim_linear(document=document,
                                           p_base=(x + self.box_width / 2,
                                                   y + self.box_height + settings["text_dim_distance_horizontal"] * 2),
                                           p1=(x, y + self.box_height),
                                           p2=(x + self.box_width, y + self.box_height),
                                           dimstyle=settings["dim_style_boxing"])
                dim_elements += dim_linear(document=document,
                                           p_base=(x - settings["text_dim_distance_vertical"],
                                                   y + self.box_height / 2),
                                           p1=(x, y),
                                           p2=(x, y + self.box_height),
                                           rotation=90,
                                           dimstyle=settings["dim_style_boxing"])

                if dimensions_inner:
                    # Polygon dimensions.
                    for i in range(len(front_points) - 1):
                        dim_elements += dim_linear(document=document,
                                                   p_base=(x + abs(front_points[i][0] - front_points[i + 1][0]) / 2,
                                                           y + self.height + settings["text_dim_distance_vertical"]),
                                                   p1=(x + front_points[i][0], y + self.height),
                                                   p2=(x + front_points[i + 1][0], y + self.height),
                                                   dimstyle=settings["dim_style_inner"])

        # Setting elements dict.
        elements["concrete_elements"] = concrete_elements
        elements["dimensions"] = dim_elements
        elements["all_elements"] = concrete_elements + dim_elements

        return elements

    def draw_transverse(self,
                        document: Drawing,
                        x: float = None,
                        y: float = None,
                        dimensions: bool = True,
                        dimensions_boxing: bool = True,
                        dimensions_inner: bool = False,
                        settings: dict = CONCRETE_SET_TRANSVERSE) -> dict:
        """
        Draws the concrete section in the transverse direction with optional dimensioning.

        :param document: The drawing document to which the concrete section will be added.
        :type document: Drawing
        :param x: X-coordinate for the placement of the concrete section. Defaults to None, in which case `self.x` is
         used.
        :type x: float, optional
        :param y: Y-coordinate for the placement of the concrete section. Defaults to None, in which case `self.y` is
         used.
        :type y: float, optional
        :param dimensions: Flag to indicate whether to draw dimensions. Defaults to True.
        :type dimensions: bool, optional
        :param dimensions_boxing: Flag to indicate whether to draw boxing dimensions. Defaults to True.
        :type dimensions_boxing: bool, optional
        :param dimensions_inner: Flag to indicate whether to draw inner dimensions. Defaults to False.
        :type dimensions_inner: bool, optional
        :param settings: Dictionary of settings for dimensioning. Defaults to `CONCRETE_SET_TRANSVERSE`.
        :type settings: dict, optional

        :return: A dictionary with keys "concrete_elements" and "all_elements", each containing the corresponding drawing
         elements.
        :rtype: dict
        """
        if x is None:
            x = self.x
        if y is None:
            y = self.y

        elements = {}
        concrete_elements = []
        dim_elements = []

        # Drawing concrete shape.
        concrete_elements += polyline(document=document, vertices=self.vertices, closed=True)

        # Moves all elements to x, y coordinates given.
        min_x = self.polygon.get_left_point()[0][0]
        min_y = self.polygon.get_bottom_point()[0][1]
        vector_translate = (x - min_x, y - min_y)
        translate(objects=concrete_elements, vector=vector_translate)

        if dimensions:
            if dimensions_boxing:
                # Boxing dimensions.
                dim_elements += dim_linear(document=document,
                                           p_base=(x + self.box_width_transverse / 2,
                                                   y + self.box_height_transverse + settings[
                                                       "text_dim_distance_horizontal"]),
                                           p1=(x, y + self.box_height_transverse),
                                           p2=(x + self.box_width_transverse, y + self.box_height_transverse),
                                           dimstyle=settings["dim_style_boxing"])
                dim_elements += dim_linear(document=document,
                                           p_base=(x - settings["text_dim_distance_vertical"],
                                                   y + self.box_height_transverse / 2),
                                           p1=(x, y),
                                           p2=(x, y + self.box_height_transverse),
                                           rotation=90,
                                           dimstyle=settings["dim_style_boxing"])

            # Inner dimensions.
            if dimensions_inner:
                sides_equations = self.polygon.get_side_equations()
                sides_centers = self.polygon.get_side_centers()

                for i, point1 in enumerate(self.vertices):
                    if i == len(self.vertices) - 1:
                        point2 = self.vertices[0]
                    else:
                        point2 = self.vertices[i + 1]
                    angle = get_angle(point1=point1, point2=point2)

                    point_base = displace_perpendicular(x0=sides_centers[i][0],
                                                        y0=sides_centers[i][1],
                                                        a=sides_equations[i][0],
                                                        d=-settings["text_dim_inner_perpendicular_distance"])
                    dim_elements += dim_linear(document=document,
                                               p_base=(point_base[0] + vector_translate[0],
                                                       point_base[1] + vector_translate[1]),
                                               p1=(point1[0] + vector_translate[0], point1[1] + vector_translate[1]),
                                               p2=(point2[0] + vector_translate[0], point2[1] + vector_translate[1]),
                                               rotation=angle,
                                               dimstyle=settings["dim_style_inner"])

        # Setting elements dict.
        elements["concrete_elements"] = concrete_elements
        elements["all_elements"] = concrete_elements + dim_elements

        return elements



# -*- coding: utf-8 -*-

# Local imports.
from etacad.globals import DRotation
from etacad.concrete import Concrete

# External imports.
import ezdxf
import pytest

from ezdxf.math import Vec3


@pytest.fixture
def concrete_length_hexagon():
    return Concrete(vertices=[(0.5, 1.7321), (1.5, 1.7321), (2, 0.866), (1.5, 0), (0.5, 0), (0, 0.866)],
                    length=4,
                    x=15,
                    y=5.5)


def test_concrete_length_hexagon_attributes(concrete_length_hexagon):
    # Geometric attributes.
    assert concrete_length_hexagon.dim3D == 4

    # Boxing attributes.
    assert concrete_length_hexagon.box_width == 4
    assert concrete_length_hexagon.box_height == 1.7321


def test_concrete_length_hexagon_draw_longitudinal(concrete_length_hexagon):
    doc = ezdxf.new(dxfversion="R2010", setup=True)
    entities = concrete_length_hexagon.draw_longitudinal(document=doc,
                                                         x=1,
                                                         y=1,
                                                         dimensions=True,
                                                         dimensions_inner=True)
    doc.saveas(filename="./tests/concrete_length_hexagon_draw_longitudinal.dxf")

    assert len(entities["concrete_elements"]) == 5
    assert len(entities["dimensions"]) == 4
    assert len(entities["all_elements"]) == 9


def test_concrete_length_hexagon_draw_transverse(concrete_length_hexagon):
    doc = ezdxf.new(dxfversion="R2010", setup=True)
    entities = concrete_length_hexagon.draw_transverse(document=doc)
    doc.saveas(filename="./tests/concrete_length_hexagon_draw_transverse.dxf")

    assert len(entities["concrete_elements"]) == 1


@pytest.fixture
def concrete_height_hexagon():
    return Concrete(vertices=[(0.5, 1.7321), (1.5, 1.7321), (2, 0.866), (1.5, 0), (0.5, 0), (0, 0.866)],
                    height=4,
                    x=15,
                    y=5.5)


def test_concrete_height_hexagon_draw_longitudinal(concrete_height_hexagon):
    doc = ezdxf.new(dxfversion="R2010", setup=True)
    entities = concrete_height_hexagon.draw_longitudinal(document=doc, dimensions=True, dimensions_inner=True)
    doc.saveas(filename="./tests/concrete_height_hexagon_draw_longitudinal.dxf")

    assert len(entities["concrete_elements"]) == 6


@pytest.fixture
def concrete_polygon():
    return Concrete(vertices=[(0.3527471624959566, 0.11004239479793698),
                              (0.216834688014635, 0.1886555176639173),
                              (0.24067898287947642, 0.3935260785542558),
                              (0.4886596369772782, 0.484050279638899),
                              (0.5196572198552669, 0.37208613616205444),
                              (0.6436475469041678, 0.21962432308779434),
                              (0.6436475469041678, 0.16245114261401516),
                              (0.5482703719078552, 0.1410112002218138),
                              (0.476737491776384, 0.22438875460363206),
                              (0.4075890380072602, 0.22438875460363206)],
                    length=4,
                    x=15,
                    y=5.5)


def test_concrete_length_polygon_draw_longitudinal(concrete_polygon):
    doc = ezdxf.new(dxfversion="R2010", setup=True)
    entities = concrete_polygon.draw_longitudinal(document=doc, dimensions=True, dimensions_inner=True)
    doc.saveas(filename="./tests/concrete_length_polygon_draw_longitudinal.dxf")

    assert len(entities["concrete_elements"]) == 10
    assert len(entities["dimensions"]) == 9
    assert len(entities["all_elements"]) == 19


def test_concrete_length_polygon_draw_transverse(concrete_polygon):
    doc = ezdxf.new(dxfversion="R2010", setup=True)
    entities = concrete_polygon.draw_transverse(document=doc, dimensions=True, dimensions_boxing=True, dimensions_inner=True)
    doc.saveas(filename="./tests/concrete_length_polygon_draw_transverse.dxf")

    assert len(entities["concrete_elements"]) == 1



# -*- coding: utf-8 -*-

# Imports.
# Local imports.

# External imports.


def expand_dictionary(dictionary) -> list:
    """
    Expands a dictionary into a list based on the dictionary's values.

    :param dictionary: A dictionary where keys represent bar numbers and values represent the quantity of each number.
    :type dictionary: dict
    :return: A list with each bar number repeated according to its value in the dictionary.
    :rtype: list
    """
    expanded_list = []
    for key, value in dictionary.items():

        expanded_list.extend([key] * value)

    return expanded_list


def gen_symmetric_list(dictionary: dict,
                       nomenclature: str = None,
                       number_init: int = None,
                       factor: float = 1) -> tuple:
    """
    Generates a symmetric list and a list of denominations based on the provided dictionary.

    :param dictionary: A dictionary where keys represent bar diameters and values represent the quantity of bars.
    :type dictionary: dict
    :param nomenclature: Prefix to use in the denomination, defaults to None.
    :type nomenclature: str, optional
    :param number_init: Initial number for the denomination, defaults to None.
    :type number_init: int, optional
    :param factor: Factor by which the bar diameter is divided, defaults to 1.
    :type factor: float, optional
    :return: A tuple containing a symmetric list of bar diameters and a corresponding list of denominations.
    :rtype: tuple
    """
    if nomenclature is None:
        nomenclature = "#"

    if number_init is None:
        n = 0
    else:
        n = number_init

    first_odd = False
    symmetryc_list = []
    denomination_list = []
    position_list = []
    quantity_list = []

    for key, value in sorted([*dictionary.items()]):
        key_factored = key / factor
        for i in range(value // 2):
            symmetryc_list.insert(0, key_factored)
            symmetryc_list.insert(len(symmetryc_list), key_factored)
            denomination_list.insert(0, "{2}{3} {0}Ø{1}".format(value, key, nomenclature, n))
            denomination_list.insert(len(denomination_list), "{2}{3} {0}Ø{1}".format(value, key, nomenclature, n))
            position_list.insert(0, "{0}{1}".format(nomenclature, n))
            position_list.insert(len(position_list), "{0}{1}".format(nomenclature, n))
            quantity_list.insert(0, value)
            quantity_list.insert(len(position_list), value)

        if is_odd(value):
            symmetryc_list.insert(len(symmetryc_list) // 2, key_factored)
            denomination_list.insert(len(denomination_list) // 2, "{2}{3} {0}Ø{1}".format(value, key, nomenclature, n))
            position_list.insert(len(position_list) // 2, "{0}{1}".format(nomenclature, n))
            quantity_list.insert(len(position_list) // 2, value)
            if first_odd:
                return ()
            else:
                first_odd = True
        n += 1

    return symmetryc_list, denomination_list, position_list, quantity_list


def gen_position_bars(dictionaries: list,
                      nomenclature: str = None,
                      number_init: int = None) -> dict:
    if nomenclature is None:
        nomenclature = "#"

    if number_init is None:
        n = 0
    else:
        n = number_init

    positions = {}
    for dict_as in dictionaries:
        for key, value in [*dict_as.items()]:
            position_name = "{0}{1}".format(nomenclature, n)
            positions[position_name] = {"diameter": key, "quantity": value}
            n += 1

    return positions


def is_odd(number):
    """
    Checks if a number is odd.

    :param number: The number to check.
    :type number: int
    :return: True if the number is odd, False otherwise.
    :rtype: bool
    """
    return bool(number % 2)


def max_per_position(*lists):
    """
    Returns a new list containing the maximum value at each position
    across multiple input lists.

    :param lists: A variable number of lists (at least two) with numerical values.
    :type lists: list of float or int
    :raises ValueError: If the lists do not all have the same length.
    :return: A list containing the maximum value at each position.
    :rtype: list

    :example:

    >>> list1 = [1.5, 3.2, 4.7, 2.9]
    >>> list2 = [2.1, 2.8, 5.0, 1.7]
    >>> list3 = [1.8, 3.3, 4.6, 2.5]
    >>> max_per_position(list1, list2, list3)
    [2.1, 3.3, 5.0, 2.9]
    """
    # Check that all lists have the same length.
    if not all(len(lst) == len(lists[0]) for lst in lists):
        raise ValueError("All lists must have the same length.")

    # Use zip to group elements by their position across all lists.
    return [max(values) for values in zip(*lists)]


def text_width_estimation(text: str, text_height: float, proportion: float = 1) -> float:
    """
    Estimate the width of a text string based on its height.

    This function approximates the width of a text string by assuming each character
    in the string occupies a width proportional to the given text height.

    :param text: The text string whose width needs to be estimated.
    :type text: str
    :param text_height: The height of the text used for the width estimation.
    :type text_height: float
    :param proportion: The proportion estimator.
    :type proportion: float
    :return: The estimated width of the text string.
    :rtype: float

    :example:

    >>> text_width_estimation("Hello", 10.0)
    37.5

    . note::
       This is a simple estimation and may not be accurate for all fonts or
       character sets. The constant `0.65` for default used for width estimation
       is based on a typical average and might need adjustment for different fonts
       or styles.
    """
    if not type(text) == str:
        try:
            text = str(text)
        except TypeError:
            raise TypeError("Text must be a string, integer or float.")

    text_width = 0
    for character in text:
        text_width += text_height * proportion

    return text_width


def str_to_dict_bar(data: str) -> dict:
    """
    Converts a string representing bars into a dictionary.

    :param data: A string where each element is in the format "quantity db diameter" and separated by "+".
    :type data: str
    :return: A dictionary where keys are bar diameters and values are the corresponding quantities.
    :rtype: dict
    """
    data_list = data.replace(" ", "").split("+")

    data_dict = {}
    for element in data_list:
        key_value = element.split("db")

        data_dict[int(key_value[1])] = int(key_value[0])

    data_dict_ordered = {k: v for k, v in sorted(data_dict.items(), key=lambda item: item[0], reverse=False)}

    return data_dict_ordered


def unpack_nested_dicts(nested_dict: dict | list[dict]):
    """
    Recursively unpacks all values from a (potentially) nested dictionary and returns them in a single list.

    :param dict nested_dict: The nested dictionary to unpack.
    :return: A list of all the values from the dictionary, including nested dictionaries.
    :rtype: list

    The function handles dictionaries with multiple levels of nesting. If a value is a dictionary, it recursively
    processes it. Non-dictionary values are directly appended to the result list.

    :example:
    >>> nested_dict = {
           'key1': {'a': 1, 'b': {'x': 7, 'y': 8}, 'c': 3},
           'key2': {'d': 4, 'e': 5},
           'key3': 6,
           'key4': {'f': {'g': 9, 'h': 10}},
           'key5': 'value'}

    >>> result = unpack_nested_dicts(nested_dict)
    [1, 7, 8, 3, 4, 5, 6, 9, 10, 'value']
    """
    print(nested_dict)
    result = []
    if isinstance(nested_dict, list) and all([isinstance(dictionary, dict) for dictionary in nested_dict]):
        for dictionary in nested_dict:
            for value in dictionary.values():
                if isinstance(value, dict):  # If the value is another dictionary, process it recursively.
                    result.extend(unpack_nested_dicts(value))
                elif isinstance(value, list):
                    result += value  # If not a dictionary, add the value directly.
                else:
                    result.append(value)
        return result

    elif isinstance(nested_dict, dict):
        for value in nested_dict.values():
            if isinstance(value, dict):  # If the value is another dictionary, process it recursively.
                result.extend(unpack_nested_dicts(value))
            elif isinstance(value, list):
                result += value  # If not a dictionary, add the value directly.
            else:
                result.append(value)
        return result

    else:
        return nested_dict



# -*- coding: utf-8 -*-

# Local imports.
from etacad.globals import Direction, Orientation, STEEL_WEIGHT
from etacad.bar import Bar

# External imports.
import ezdxf
import pytest

from ezdxf.math import Vec3
from math import pi


@pytest.fixture
def bar_straight_horizontal():
    return Bar(reinforcement_length=12,
               diameter=0.012,
               x=1,
               y=1,
               left_anchor=0,
               right_anchor=0,
               mandrel_radius=0,
               direction=Direction.HORIZONTAL,
               orientation=Orientation.BOTTOM,
               bend_longitud=0,
               bend_angle=0,
               bend_height=0,
               denomination="@bar_straight_horizontal")


def test_attributes_bar_straight_horizontal(bar_straight_horizontal):
    # Bending attributes.
    assert bar_straight_horizontal.bend_longitud == 0
    assert bar_straight_horizontal.bend_angle == 0
    assert bar_straight_horizontal.bend_height == 0
    assert bar_straight_horizontal.bending_proyection == 0

    # Mandrel attributes.
    assert bar_straight_horizontal.left_anchor == 0
    assert bar_straight_horizontal.right_anchor == 0
    assert bar_straight_horizontal.mandrel_radius == 0
    assert bar_straight_horizontal.mandrel_radius_ext == 0.012

    # Geometric attributes.
    assert bar_straight_horizontal.diameter == 0.012
    assert bar_straight_horizontal.radius == 0.006
    assert bar_straight_horizontal.direction == Direction.HORIZONTAL
    assert bar_straight_horizontal.orientation == Orientation.BOTTOM
    assert bar_straight_horizontal.transverse_center is None
    assert bar_straight_horizontal.reinforcement_length == 12
    assert bar_straight_horizontal.length == 12
    assert bar_straight_horizontal.weight == ((0.012 ** 2) * pi / 4) * 12 * STEEL_WEIGHT

    # Boxing attributes.
    assert bar_straight_horizontal.box_width == 12
    assert bar_straight_horizontal.box_height == 0.012

    # Others.
    assert bar_straight_horizontal.denomination == "@bar_straight_horizontal"


def test_draw_longitudinal_bar_straight_horizontal(bar_straight_horizontal):
    doc = ezdxf.new(dxfversion="R2010", setup=True)

    ex_01 = bar_straight_horizontal.draw_longitudinal(document=doc, x=2, y=1, unifilar=False, dimensions=False)
    assert len(ex_01["all_elements"]) == 5
    assert ex_01["steel_elements"][0].dxf.start == Vec3(2, 1.012, 0)  # Top side start.
    assert ex_01["steel_elements"][0].dxf.end == Vec3(14, 1.012, 0)  # Top side end.
    assert ex_01["steel_elements"][1].dxf.start == Vec3(2, 1, 0)  # Bottom side start.
    assert ex_01["steel_elements"][1].dxf.end == Vec3(14, 1, 0)  # Bottom side end.

    ex_02 = bar_straight_horizontal.draw_longitudinal(document=doc, x=2, y=0, unifilar=False, denomination=False)
    assert len(ex_02["all_elements"]) == 5
    assert ex_02["steel_elements"][0].dxf.start == Vec3(2, 0.012, 0)  # Top side start.
    assert ex_02["steel_elements"][0].dxf.end == Vec3(14, 0.012, 0)  # Top side end.
    assert ex_02["steel_elements"][1].dxf.start == Vec3(2, 0, 0)  # Bottom side start.
    assert ex_02["steel_elements"][1].dxf.end == Vec3(14, 0, 0)  # Bottom side end.

    ex_03 = bar_straight_horizontal.draw_longitudinal(document=doc, x=2, y=-1, unifilar=False)
    assert len(ex_03["all_elements"]) == 6
    assert ex_03["steel_elements"][0].dxf.start == Vec3(2, -0.988, 0)  # Top side start.
    assert ex_03["steel_elements"][0].dxf.end == Vec3(14, -0.988, 0)  # Top side end.
    assert ex_03["steel_elements"][1].dxf.start == Vec3(2, -1, 0)  # Bottom side start.
    assert ex_03["steel_elements"][1].dxf.end == Vec3(14, -1, 0)  # Bottom side end.

    ex_04 = bar_straight_horizontal.draw_longitudinal(document=doc, x=-12, y=1, unifilar=False, dimensions=False)
    assert len(ex_04["all_elements"]) == 5
    assert ex_04["steel_elements"][0].dxf.start == Vec3(-12, 1.012, 0)  # Top side start.
    assert ex_04["steel_elements"][0].dxf.end == Vec3(0, 1.012, 0)  # Top side end.
    assert ex_04["steel_elements"][1].dxf.start == Vec3(-12, 1, 0)  # Bottom side start.
    assert ex_04["steel_elements"][1].dxf.end == Vec3(0, 1, 0)  # Bottom side end.

    ex_05 = bar_straight_horizontal.draw_longitudinal(document=doc, x=-12, y=0, unifilar=False, denomination=False)
    assert len(ex_05["all_elements"]) == 5
    assert ex_05["steel_elements"][0].dxf.start == Vec3(-12, 0.012, 0)  # Top side start.
    assert ex_05["steel_elements"][0].dxf.end == Vec3(0, 0.012, 0)  # Top side end.
    assert ex_05["steel_elements"][1].dxf.start == Vec3(-12, 0, 0)  # Bottom side start.
    assert ex_05["steel_elements"][1].dxf.end == Vec3(0, 0, 0)  # Bottom side end.

    ex_06 = bar_straight_horizontal.draw_longitudinal(document=doc, x=-12, y=-1, unifilar=False)
    assert len(ex_06["all_elements"]) == 6
    assert ex_06["steel_elements"][0].dxf.start == Vec3(-12, -0.988, 0)  # Top side start.
    assert ex_06["steel_elements"][0].dxf.end == Vec3(0, -0.988, 0)  # Top side end.
    assert ex_06["steel_elements"][1].dxf.start == Vec3(-12, -1, 0)  # Bottom side start.
    assert ex_06["steel_elements"][1].dxf.end == Vec3(0, -1, 0)  # Bottom side end.

    ex_07 = bar_straight_horizontal.draw_longitudinal(document=doc, x=2, y=2, unifilar=True)
    assert len(ex_07["all_elements"]) == 3
    assert ex_07["steel_elements"][0].dxf.start == Vec3(2, 2, 0)  # Top side start.
    assert ex_07["steel_elements"][0].dxf.end == Vec3(14, 2, 0)  # Top side end.
    assert ex_07["steel_elements"][0].dxf.end[0] - ex_07["steel_elements"][0].dxf.start[0] == 12

    ex_08 = bar_straight_horizontal.draw_longitudinal(document=doc, x=-12, y=2, unifilar=True)
    assert len(ex_08["all_elements"]) == 3
    assert ex_08["steel_elements"][0].dxf.start == Vec3(-12, 2, 0)
    assert ex_08["steel_elements"][0].dxf.end == Vec3(0, 2, 0)
    assert ex_08["steel_elements"][0].dxf.end[0] - ex_08["steel_elements"][0].dxf.start[0] == 12

    doc.saveas(filename="./tests/bar_straight_horizontal.dxf")

    # General.
    entities = (ex_01["all_elements"] + ex_02["all_elements"] + ex_03["all_elements"] + ex_04["all_elements"] +
                ex_05["all_elements"] + ex_06["all_elements"] + ex_07["all_elements"] + ex_08["all_elements"])
    assert len(entities) == 38


def test_extract_data_bar_straight_horizontal(bar_straight_horizontal):
    assert bar_straight_horizontal.extract_data(["diameter", "length"]) == [0.012, 12]


@pytest.fixture
def bar_straight_vertical():
    return Bar(reinforcement_length=12,
               diameter=0.01,
               x=1,
               y=1,
               left_anchor=0,
               right_anchor=0,
               mandrel_radius=0,
               direction=Direction.VERTICAL,
               orientation=Orientation.BOTTOM,
               bend_longitud=0,
               bend_angle=0,
               bend_height=0,
               denomination="@bar_straight_vertical")


def test_attributes_bar_straight_vertical(bar_straight_vertical):
    # Bending attributes.
    assert bar_straight_vertical.bend_longitud == 0
    assert bar_straight_vertical.bend_angle == 0
    assert bar_straight_vertical.bend_height == 0
    assert bar_straight_vertical.bending_proyection == 0

    # Mandrel attributes.
    assert bar_straight_vertical.left_anchor == 0
    assert bar_straight_vertical.right_anchor == 0
    assert bar_straight_vertical.mandrel_radius == 0
    assert bar_straight_vertical.mandrel_radius_ext == 0.01

    # Geometric attributes.
    assert bar_straight_vertical.diameter == 0.01
    assert bar_straight_vertical.radius == 0.005
    assert bar_straight_vertical.direction == Direction.VERTICAL
    assert bar_straight_vertical.orientation == Orientation.BOTTOM
    assert bar_straight_vertical.transverse_center is None
    assert bar_straight_vertical.reinforcement_length == 12
    assert bar_straight_vertical.length == 12
    assert bar_straight_vertical.weight == ((0.01 ** 2) * pi / 4) * 12 * STEEL_WEIGHT

    # Boxing attributes.
    assert bar_straight_vertical.box_width == 12
    assert bar_straight_vertical.box_height == 0.01

    # Others.
    assert bar_straight_vertical.denomination == "@bar_straight_vertical"


def test_draw_longitudinal_bar_straight_vertical(bar_straight_vertical):
    doc = ezdxf.new(dxfversion="R2010", setup=True)

    ex_01 = bar_straight_vertical.draw_longitudinal(document=doc, x=-2, y=1, unifilar=False, dimensions=False)
    assert len(ex_01["all_elements"]) == 5
    assert ex_01["steel_elements"][0].dxf.start == Vec3(-2, 1, 0)  # Top side start.
    assert ex_01["steel_elements"][0].dxf.end == Vec3(-1.9999999999999991, 13, 0)  # Top side end.
    assert ex_01["steel_elements"][1].dxf.start == Vec3(-1.99, 1, 0)  # Bottom side start.
    assert ex_01["steel_elements"][1].dxf.end == Vec3(-1.989999999999999, 13, 0)  # Bottom side end.

    ex_02 = bar_straight_vertical.draw_longitudinal(document=doc, x=-1, y=1, unifilar=False, denomination=False)
    assert len(ex_02["all_elements"]) == 5
    assert ex_02["steel_elements"][0].dxf.start == Vec3(-1, 1, 0)  # Top side start.
    assert ex_02["steel_elements"][0].dxf.end == Vec3(-0.9999999999999993, 13, 0)  # Top side end.
    assert ex_02["steel_elements"][1].dxf.start == Vec3(-0.99, 1, 0)  # Bottom side start.
    assert ex_02["steel_elements"][1].dxf.end == Vec3(-0.9899999999999993, 13, 0)  # Bottom side end.

    ex_03 = bar_straight_vertical.draw_longitudinal(document=doc, x=0, y=1, unifilar=False)
    assert len(ex_03["all_elements"]) == 6
    assert ex_03["steel_elements"][0].dxf.start == Vec3(0, 1, 0)  # Top side start.
    assert ex_03["steel_elements"][0].dxf.end == Vec3(6.661338147750939e-16, 13, 0)  # Top side end.
    assert ex_03["steel_elements"][1].dxf.start == Vec3(0.010000000000000009, 1, 0)  # Bottom side start.
    assert ex_03["steel_elements"][1].dxf.end == Vec3(0.010000000000000786, 13, 0)  # Bottom side end.

    ex_04 = bar_straight_vertical.draw_longitudinal(document=doc, x=1, y=1, unifilar=True, dimensions=False)
    assert len(ex_04["all_elements"]) == 2
    assert ex_04["steel_elements"][0].dxf.start == Vec3(0.9999999999999998, 1, 0)  # Top side start.
    assert ex_04["steel_elements"][0].dxf.end == Vec3(1.0000000000000007, 13, 0)  # Top side end.

    ex_05 = bar_straight_vertical.draw_longitudinal(document=doc, x=2, y=1, unifilar=True)
    assert len(ex_05["all_elements"]) == 3
    assert ex_05["steel_elements"][0].dxf.start == Vec3(1.9999999999999998, 1, 0)  # Top side start.
    assert ex_05["steel_elements"][0].dxf.end == Vec3(2.000000000000001, 13, 0)  # Top side end.

    # General.
    entities = (ex_01["all_elements"] + ex_02["all_elements"] + ex_03["all_elements"] + ex_04["all_elements"] +
                ex_05["all_elements"])
    assert len(entities) == 21

    doc.saveas(filename="./tests/bar_straight_vertical.dxf")


@pytest.fixture
def bar_straight_horizontal_lab():
    return Bar(reinforcement_length=4,
               diameter=0.01,
               x=0,
               y=0,
               left_anchor=0.15,
               right_anchor=0,
               mandrel_radius=0.01,
               direction=Direction.HORIZONTAL,
               orientation=Orientation.BOTTOM,
               bend_longitud=0,
               bend_angle=0,
               bend_height=0,
               denomination="bar_straight_vertical_lab")


def test_attributes_straight_horizontal_lab(bar_straight_horizontal_lab):
    # Bending attributes.
    assert bar_straight_horizontal_lab.bend_longitud == 0
    assert bar_straight_horizontal_lab.bend_angle == 0
    assert bar_straight_horizontal_lab.bend_height == 0
    assert bar_straight_horizontal_lab.bending_proyection == 0

    # Mandrel attributes.
    assert bar_straight_horizontal_lab.left_anchor == 0.15
    assert bar_straight_horizontal_lab.right_anchor == 0
    assert bar_straight_horizontal_lab.mandrel_radius == 0.01
    assert bar_straight_horizontal_lab.mandrel_radius_ext == 0.02

    # Geometric attributes.
    assert bar_straight_horizontal_lab.diameter == 0.01
    assert bar_straight_horizontal_lab.radius == 0.005
    assert bar_straight_horizontal_lab.direction == Direction.HORIZONTAL
    assert bar_straight_horizontal_lab.orientation == Orientation.BOTTOM
    assert bar_straight_horizontal_lab.transverse_center is None
    assert bar_straight_horizontal_lab.reinforcement_length == 4
    assert bar_straight_horizontal_lab.length == 4.15
    assert bar_straight_horizontal_lab.weight == ((0.01 ** 2) * pi / 4) * 4.15 * STEEL_WEIGHT

    # Boxing attributes.
    assert bar_straight_horizontal_lab.box_width == 4
    assert bar_straight_horizontal_lab.box_height == 0.16999999999999998

    # Others.
    assert bar_straight_horizontal_lab.denomination == "bar_straight_vertical_lab"


@pytest.fixture
def bar_straight_hotizontal_rab():
    pass


@pytest.fixture
def bar_straight_horizontal_lab_rab():
    pass



# -*- coding: utf-8 -*-

# Local imports.
from etacad.cadtable import CADTable

# External imports.
import ezdxf
import pytest

from ezdxf.math import Vec3


@pytest.fixture
def cadtable_rows_columns():
    return CADTable(rows=10,
                    columns=11,
                    labels=["c01", "c10", "c22", "c43", "c64"],
                    index=[*range(11)],
                    x=5,
                    y=6)


def test_attributes_cadtable_rows_columns(cadtable_rows_columns):
    pass


def test_draw_table_cadtable_rows_columns(cadtable_rows_columns):
    doc = ezdxf.new(dxfversion="R2010", setup=True)
    content_grid = cadtable_rows_columns.draw_grid(document=doc)
    labels_grid = cadtable_rows_columns.draw_labels(document=doc)
    doc.saveas(filename="./tests/cadtable_draw_table_rows_columns.dxf")

    # Stirrups.

    # General.
    assert len(content_grid["grid_hz_lines"]) == 11
    assert len(content_grid["grid_vt_lines"]) == 12


@pytest.fixture
def cadtable_data():
    return CADTable(data=[["1", "16", "-", "4", "None"],
                          ["2", "12", "-", "2", "None"],
                          ["4", "8", "-", "22", "None"]],
                    labels=["POSITION", "DB", "SEP", "QUANTITY", "GRAPH"])


def test_attributes_cadtable_data(cadtable_data):

    # General attributes.
    assert cadtable_data.rows == 3
    assert cadtable_data.columns == 5


def test_draw_grid_cadtable_data(cadtable_data):
    doc = ezdxf.new(dxfversion="R2010", setup=True)
    content_grid = cadtable_data.draw_grid(document=doc)
    labels_grid = cadtable_data.draw_labels(document=doc)
    doc.saveas(filename="./tests/cadtable_draw_grid_data.dxf")

    # General.
    assert len(content_grid["grid_hz_lines"]) == 4
    assert len(content_grid["grid_vt_lines"]) == 6


def test_draw_table_cadtable_data(cadtable_data):
    doc = ezdxf.new(dxfversion="R2010", setup=True)
    table = cadtable_data.draw_table(document=doc, x=1, y=1)
    doc.saveas(filename="./tests/cadtable_draw_table_data.dxf")

    # General.
    assert len(table["labels"]["texts"]) == 5
    assert len(table["content"]["texts_row0"]) == 5
    assert len(table["content"]["texts_row1"]) == 5
    assert len(table["content"]["texts_row2"]) == 5



# -*- coding: utf-8 -*-

"""A package aimed to simpilfy drawings of structurals elements based on ezdxf library."""

__version__ = "0.0.11"

from .bar import Bar
from .stirrup import Stirrup
from .concrete import Concrete
from .cadtable import CADTable
from .beam import Beam
from .column import Column



# -*- coding: utf-8 -*-

# Local imports.
from etacad.globals import ColumnTypes, Direction, Orientation
from etacad.column import Column

# External imports.
import ezdxf
import pytest

from ezdxf.math import Vec3


@pytest.fixture
def column_concrete():
    return Column(width=0.8,
                  depth=0.8,
                  height=6,
                  x=-1,
                  y=-1)


def test_attributes_column_concrete(column_concrete):
    # Geometric attributes.
    assert column_concrete.width == .8
    assert column_concrete.depth == .8
    assert column_concrete.height == 6
    assert column_concrete.x == -1
    assert column_concrete.y == -1

    # Concrete attributes.
    assert column_concrete.concrete.volume == .8 * .8 * 6


def test_draw_longitudinal_column_concrete(column_concrete):
    doc = ezdxf.new(dxfversion="R2010", setup=True)
    entities = column_concrete.draw_longitudinal(document=doc, unifilar_bars=False)
    doc.saveas(filename="./tests/column_concrete_longitudinal.dxf")

    # Concrete.
    assert len(entities["concrete"]["concrete_elements"]) == 4

    # Dimensions.
    assert len(entities["dimensions_elements"]) == 0


def test_draw_transverse_column_concrete(column_concrete):
    doc = ezdxf.new(dxfversion="R2010", setup=True)
    entities = column_concrete.draw_transverse(document=doc)
    doc.saveas(filename="./tests/column_concrete_transverse.dxf")

    # Concrete.
    assert len(entities["concrete"]["concrete_elements"]) == 1


@pytest.fixture
def column():
    return Column(width=.2,
                  depth=.2,
                  height=6,
                  cover=.03,
                  x=20,
                  y=10,
                  as_sup={0.016: 2, 0.012: 2},
                  as_right={0.012: 2},
                  as_inf={0.016: 2, 0.012: 2},
                  as_left={0.012: 2},
                  stirrups_db=[.006, 0.008, .006, .006, 0.008, .006],
                  stirrups_anchor=[.1, .12, .1, .1, .12, .1],
                  stirrups_sep=[.1, .18, .1, .1, .18, .1],
                  stirrups_length=[.6, 1.3, .6, .6, 1.3, .6],
                  stirrups_x=[.1, .85, 2.3, 3.1, 3.85, 5.3],
                  beams=[[.2, .4], [.2, .3], [.2, .3]],
                  beams_pos=[.2, 3, 5.7],
                  beams_symbol=["B1", "B2", "C1"])


def test_attributes_rectangular_column(column):
    # Geometric attributes.
    assert column.width == .2
    assert column.depth == .2
    assert column.height == 6
    assert column.diameter is None
    assert column.column_type == ColumnTypes.RECTANGULAR
    assert column.x == 20
    assert column.y == 10
    assert column.cover == .03
    assert column.direction == Direction.VERTICAL
    assert column.orientation == Orientation.RIGHT

    # Longitudinal steel attributes.
    assert column.as_sup == {0.012: 2, 0.016: 2}
    assert column.as_right == {0.012: 2}
    assert column.as_inf == {0.012: 2, 0.016: 2}
    assert column.as_left == {0.012: 2}
    assert column.anchor_sup == [0, 0, 0, 0]
    assert column.anchor_right == [0, 0]
    assert column.anchor_inf == [0, 0, 0, 0]
    assert column.anchor_left == [0, 0]

    assert column.max_db_sup == 0.016
    assert column.max_db_right == 0.012
    assert column.max_db_inf == 0.016
    assert column.max_db_left == 0.012

    # Box attributes.
    assert column.box_width == .2
    assert column.box_height == 6

    # Stirrup dimensions.
    stirrup_width = (column.width - column.cover * 2 + column.max_db_inf + 0.006 * 2)
    assert stirrup_width == .16800000000000004
    assert column.stirrups[0].width == .16800000000000004


def test_draw_longitudinal_rectangular_column(column):
    doc = ezdxf.new(dxfversion="R2010", setup=True)
    entities = column.draw_longitudinal(document=doc, x=2, y=3, unifilar_bars=False)
    doc.saveas(filename="./tests/column_rectangular_longitudinal.dxf")

    # Bars.
    assert entities["bars"][0]["steel_elements"][0].dxf.start == Vec3(2.0219999999999985, 3.0299999999999994, 0.0)

    # General.
    assert len(entities["concrete"]["all_elements"]) == 6
    assert len(entities["all_elements"]) == 90


def test_draw_longitudinal_unifilar_rectangular_column(column):
    doc = ezdxf.new(dxfversion="R2010", setup=True)
    entities = column.draw_longitudinal(document=doc, x=2, y=3, unifilar_bars=True)
    doc.saveas(filename="./tests/column_rectangular_unifilar.dxf")

    # Stirrups.
    stirrup_width = (column.width - column.cover * 2 + column.max_db_inf + 0.006 * 2)
    assert stirrup_width == .16800000000000004
    assert column.stirrups[0].width == .16800000000000004

    # General.
    assert len(entities["all_elements"]) == 78


def test_list_to_stirrups_rectangular_column(column):
    pass


def test_draw_transverse_rectangular_column(column):
    doc = ezdxf.new(dxfversion="R2010", setup=True)
    entities_st_1 = column.draw_transverse(document=doc, x=0.5, y=1, y_section=0.65)
    entities_st_2 = column.draw_transverse(document=doc, x=-0.5, y=1, y_section=1.5)
    entities_st_3 = column.draw_transverse(document=doc, x=-0.5, y=-1, y_section=3.65)
    entities_st_4 = column.draw_transverse(document=doc, x=0.5, y=-1, y_section=4.5)
    doc.saveas(filename="./tests/column_rectangular_transverse.dxf")

    # General.
    assert len(entities_st_1["all_elements"]) == 37
    assert len(entities_st_2["all_elements"]) == 37
    assert len(entities_st_3["all_elements"]) == 37
    assert len(entities_st_4["all_elements"]) == 37


def test_draw_longitudinal_rebar_detailing_rectangular_column(column):
    doc = ezdxf.new(dxfversion="R2010", setup=True)
    entities = column.draw_longitudinal_rebar_detailing(document=doc, x=-10, y=1, unifilar=False)
    doc.saveas(filename="./tests/column_recatangular_longitudinal_rebar_detailing.dxf")

    assert entities["text_elements"][0].dxf.align_point == Vec3(-9.98, 7.6, 0.0)
    assert entities["text_elements"][0].dxf.text == "As sup."
    assert entities["barline_elements"][0].dxf.start == Vec3(-9.072, 0.7, 0.0)

    # General.
    assert len(entities["text_elements"]) == 4
    assert len(entities["bars"]) == 6
    assert len(entities["barline_elements"]) == 3
    assert len(entities["all_elements"]) == 43


def test_draw_transverse_rebar_detailing_recutangular_column(column):
    doc = ezdxf.new(dxfversion="R2010", setup=True)
    entities = column.draw_transverse_rebar_detailing(document=doc, x=-10, y=1, y_section=1)
    doc.saveas(filename="./tests/column_recatangular_transverse_rebar_detailing.dxf")

    assert entities["stirrups"][0]["steel_elements"][0].dxf.start == Vec3(-9.992, 1.0159999999999982, 0.0)

    # General.
    assert len(entities["all_elements"]) == 26


def test_draw_table_rebar_detailing_column(column):
    doc = ezdxf.new(dxfversion="R2010", setup=True)
    entities = column.draw_table_rebar_detailing(document=doc, x=-20, y=-5)
    doc.saveas(filename="./tests/column_recatangular_table_rebar_detailing.dxf")

    # General.
    assert len(entities["all_elements"]) == 122



from setuptools import setup, find_packages

readme = open("./README.md", "r")

setup(name="etacad",
      packages=find_packages(),
      version="0.0.11",
      description="A package aimed to simpilfy drawings of structurals elements based on ezdxf library.",
      long_description=readme.read(),
      long_description_content_type="text/markdown",
      author="Kevin Axel Tagliaferri",
      author_email='kevinaxeltagliaferri@hotmail.com',
      url="https://github.com/AxelTAG/etacad.git",
      install_requires=["attrs", "ezdxf"])



# -*- coding: utf-8 -*-

# Imports.
# Local imports.
from etacad.drawing_utils import delimit_axe, dim_linear, rect, text
from etacad.bar import Bar
from etacad.cadtable import CADTable
from etacad.concrete import Concrete
from etacad.globals import (BEAM_SET_LONG, BEAM_SET_LONG_REBAR, BEAM_SET_TRANSVERSE, CONCRETE_WEIGHT, Direction,
                            ElementTypes, Orientation)
from etacad.stirrup import Stirrup
from etacad.utils import gen_symmetric_list, gen_position_bars

# External imports.
from attrs import define, field
from ezdxf.document import Drawing
from ezdxf.gfxattribs import GfxAttribs
from itertools import chain


@define
class Beam:
    """
    Beam element, computes geometrics and physics props and manages dxf drawing methods (longitudinal, transversal,
    reinforcement detailing, etc.)

    :param width: Width of the beam.
    :type width: float
    :param height: Height of the beam.
    :type height: float
    :type length: Length of the beam.
    :param x: X-coordinate of the bottom left corner of the beam concrete shape.
    :type x: float
    :param y: Y-coordinate of the bottom left corner of the beam concrete shape.
    :type y: float
    :param direction: Direction of the beam (HORIZONTAL or VERTICAL).
    :type direction: str
    :param orientation: Orientation of the beam (TOP, RIGHT, DOWN, LEFT).
    :type orientation: str
    :param as_sup: Dictionary containing reinforcement information for the top side.
    :type as_sup: dict
    :param anchor_sup: Anchor length for the top reinforcement or list of lengths.
    :type anchor_sup: float | list
    :param as_right: Dictionary containing reinforcement information for the right side.
    :type as_right: dict
    :param anchor_right: Anchor length for the right reinforcement or list of lengths.
    :type anchor_right: float | list
    :param as_inf: Dictionary containing reinforcement information for the bottom side.
    :type as_inf: dict
    :param anchor_inf: Anchor length for the bottom reinforcement or list of lengths.
    :type anchor_inf: float | list
    :param as_left: Dictionary containing reinforcement information for the left side.
    :type as_left: dict
    :param anchor_left: Anchor length for the left reinforcement or list of lengths.
    :type anchor_left: float | list
    :param cover: Concrete cover for the reinforcement.
    :type cover: float
    :param concrete_specific_weight: Specific weight of the concrete. Defaults to CONCRETE_WEIGHT.
    :param stirrups_db: Diameter of stirrups or list of diameters.
    :type stirrups_db: float | list
    :param stirrups_sep: Separation between stirrups or list of separations.
    :type stirrups_sep: float | list
    :param stirrups_length: List of lengths for stirrups.
    :type stirrups_length: list
    :param stirrups_anchor: Anchor length of stirrups or list of anchor lengths.
    :type stirrups_anchor: float | list
    :param stirrups_x: List of x-coordinates for stirrups.
    :type stirrups_x: list
    :param columns: List of lists containing column width and height.
    :type columns: list[list[float, float]]
    :param columns_pos: List of positions (X-coordinate of start) for columns.
    :type columns_pos: list[float]
    :param columns_symbol: List of symbols representing columns.
    :type columns_symbol: list[str]
    :param nomenclature: Nomenclature prefix used for labeling elements.
    :type nomenclature: str
    :param number_init: Initial number for labeling elements.
    :type number_init: int
    :param element_type: Type of the structural element (e.g., BEAM).
    :type element_type: ElementTypes

    :ivar width: Width of the beam.
    :ivar height: Height of the beam.
    :ivar length: Length of the beam.
    :ivar x: X-coordinate of the bottom left corner of the beam concrete shape.
    :ivar y: Y-coordinate of the bottom left corner of the beam concrete shape.
    :ivar direction: Direction of the element (HORIZONTAL or VERTICAL).
    :ivar orientation: Orientation of the element (e.g., TOP, RIGHT, DOWN, LEFT).

    :ivar as_sup: Dictionary containing reinforcement information for the top side.
    :ivar max_db_sup: Maximum diameter of the top reinforcement bars.
    :ivar anchor_sup: Anchor length for the top reinforcement or list of lengths.
    :ivar number_init_sup: Initial numbering for the top reinforcement bars.
    :ivar as_right: Dictionary containing reinforcement information for the right side.
    :ivar max_db_right: Maximum diameter of the right reinforcement bars.
    :ivar anchor_right: Anchor length for the right reinforcement or list of lengths.
    :ivar number_init_right: Initial numbering for the right reinforcement bars.
    :ivar as_inf: Dictionary containing reinforcement information for the bottom side.
    :ivar max_db_inf: Maximum diameter of the bottom reinforcement bars.
    :ivar anchor_inf: Anchor length for the bottom reinforcement or list of lengths.
    :ivar number_init_inf: Initial numbering for the bottom reinforcement bars.
    :ivar as_left: Dictionary containing reinforcement information for the left side.
    :ivar max_db_left: Maximum diameter of the left reinforcement bars.
    :ivar anchor_left: Anchor length for the left reinforcement or list of lengths.
    :ivar number_init_left: Initial numbering for the left reinforcement bars.
    :ivar cover: Concrete cover for the reinforcement.

    :ivar concrete_specific_weight: Specific weight of the concrete.
    :ivar concrete: Concrete instance class.

    :ivar stirrups_db: Diameter of stirrups or list of diameters.
    :ivar stirrups_sep: Separation between stirrups or list of separations.
    :ivar stirrups_length: List of lengths for stirrups.
    :ivar stirrups_anchor: Anchor length of stirrups or list of anchor lengths.
    :ivar stirrups_x: List of x-coordinates for stirrups.

    :ivar columns: List of lists containing column width and height.
    :ivar columns_pos: List of positions (X-coordinate of start) for columns.
    :ivar columns_symbol: List of symbols representing columns.

    :ivar bars_as_sup: List of top reinforcement bars as entities.
    :ivar bars_as_right: List of right reinforcement bars as entities.
    :ivar bars_as_inf: List of bottom reinforcement bars as entities.
    :ivar bars_as_left: List of left reinforcement bars as entities.

    :ivar all_bars: List of all reinforcement bars as entities.
    :ivar stirrups: List of stirrups as entities.
    :ivar all_elements: List of all elements/entities in the structure.

    :ivar nomenclature: Nomenclature prefix used for labeling elements.
    :ivar number_init: Initial number for labeling elements.
    :ivar element_type: Type of the structural element (e.g., BEAM).
    """
    # Geometric attributes.
    width: float = field(converter=float)
    height: float = field(converter=float)
    length: float = field(converter=float)
    x: float = field(default=0)
    y: float = field(default=0)
    direction: Direction = field(default=Direction.HORIZONTAL)
    orientation: Orientation = field(default=Orientation.RIGHT)

    # Steel attributes.
    number_init: int = field(default=None)

    as_sup: dict = field(default=None)
    max_db_sup: float = field(init=False)
    anchor_sup: float | list = field(default=0)
    number_init_sup: int = field(init=False)

    as_right: dict = field(default=None)
    max_db_right: float = field(init=False)
    anchor_right: float | list = field(default=0)
    number_init_right: int = field(init=False)

    as_inf: dict = field(default=None)
    max_db_inf: float = field(init=False)
    anchor_inf: float | list = field(default=0)
    number_init_inf: int = field(init=False)

    as_left: dict = field(default=None)
    max_db_left: float = field(init=False)
    anchor_left: float | list = field(default=0)
    number_init_left: int = field(init=False)

    cover: float = field(default=0.025, converter=float)

    # Concrete attribute.
    concrete_specific_weight = field(default=CONCRETE_WEIGHT)
    concrete: Concrete = field(init=False)

    # Stirrups attribute.
    stirrups_db: float | list = field(default=None)
    stirrups_sep: float | list = field(default=None)
    stirrups_length: list = field(default=None)
    stirrups_anchor: float | list = field(default=None)
    stirrups_x: list = field(default=None)

    # Columns attributes.
    columns: list[list[float, float]] = field(default=None)
    columns_pos: list[float] = field(default=None)
    columns_symbol: list[str] = field(default=None)

    # Elements/Entities.
    bars_as_sup: list = field(init=False)
    bars_as_right: list = field(init=False)
    bars_as_inf: list = field(init=False)
    bars_as_left: list = field(init=False)
    all_bars: list = field(init=False)
    stirrups: list = field(init=False)
    all_elements: list = field(init=False)

    # Position bar attributes.
    nomenclature: str = field(default="#")
    positions: dict = field(init=False)

    # Beam attributes.
    denomination: str = field(default=None)
    element_type: ElementTypes = field(default=ElementTypes.BEAM)

    def __attrs_post_init__(self):
        # Longitudinal steel attributes.
        if self.number_init is None:
            self.number_init = 0

        # Top bars.
        if self.as_sup:
            self.max_db_sup = max(self.as_sup.keys())
            self.number_init_sup = self.number_init if self.number_init is not None else 0
            self.number_init = self.number_init + len(self.as_sup) if self.number_init else len(self.as_sup)

            if not type(self.anchor_sup) == list:
                self.anchor_sup = [self.anchor_sup] * sum(self.as_sup.values())
        else:
            self.as_sup, self.max_db_sup, self.number_init_sup = {}, 0, 0

        # Right bars.
        if self.as_right:
            self.max_db_right = max(self.as_right.keys())
            self.number_init_right = self.number_init if self.number_init is not None else 0
            self.number_init = self.number_init + len(self.as_right) if self.number_init else len(self.as_right)

            if not type(self.anchor_right) == list:
                self.anchor_right = [self.anchor_right] * sum(self.as_right.values())
        else:
            self.as_right, self.max_db_right, self.number_init_right = {}, 0, 0

        # Inferior bars.
        if self.as_inf:
            self.max_db_inf = max(self.as_inf.keys())
            self.number_init_inf = self.number_init if self.number_init is not None else 0
            self.number_init = self.number_init + len(self.as_inf) if self.number_init else len(self.as_inf)

            if not type(self.anchor_inf) == list:
                self.anchor_inf = [self.anchor_inf] * sum(self.as_inf.values())
        else:
            self.as_inf, self.max_db_inf, self.number_init_inf = {}, 0, 0

        # Left bars.
        if self.as_left:
            self.max_db_left = max(self.as_left.keys())
            self.number_init_left = self.number_init if self.number_init is not None else 0
            self.number_init = self.number_init + len(self.as_left) if self.number_init else len(self.as_left)

            if not type(self.anchor_left) == list:
                self.anchor_left = [self.anchor_left] * sum(self.as_left.values())
        else:
            self.as_left, self.max_db_left, self.number_init_left = {}, 0, 0

        # Concrete attributes.
        vertices = [(0, 0),
                    (0, self.height),
                    (self.width, self.height),
                    (self.width, 0)]

        self.concrete = Concrete(vertices=vertices,
                                 length=self.length,
                                 x=self.x,
                                 y=self.y,
                                 specific_weight=CONCRETE_WEIGHT)

        # Stirrups attributes.
        if self.stirrups_db is not None:
            if type(self.stirrups_db) == float:
                self.stirrups_db = [self.stirrups_db]

            if self.stirrups_anchor is None:
                self.stirrups_anchor = [0.1] * len(self.stirrups_db)

            elif type(self.stirrups_anchor) == float:
                self.stirrups_anchor = [self.stirrups_anchor] * len(self.stirrups_db)

            if type(self.stirrups_sep) == float:
                self.stirrups_sep = [self.stirrups_sep] * len(self.stirrups_db)

            if self.stirrups_length is None:
                self.stirrups_length = [self.length - (self.height + self.cover) * 2]

            if self.stirrups_x is None:
                self.stirrups_x = [self.height + self.cover]

            self.stirrups = self.__list_to_stirrups()

        else:
            self.stirrups = []

        # Column attributes.
        if self.columns is not None:
            if self.columns_symbol is None:
                self.columns_symbol = [*range(1, len(self.columns) + 1)]

        # Position bar attributes.
        min_number_init = min(self.number_init_sup, self.number_init_right, self.number_init_inf, self.number_init_left)

        self.positions = gen_position_bars(dictionaries=[self.as_sup, self.as_right, self.as_inf, self.as_left],
                                           nomenclature=self.nomenclature,
                                           number_init=min_number_init)

        # Elements/Entities.
        self.bars_as_sup = self.__dict_to_bars(self.as_sup, width=self.width, x=self.x, y=self.y, side=0,
                                               anchor=self.anchor_sup, nomenclature=self.nomenclature,
                                               number_init=self.number_init_sup) if self.as_sup else []
        self.bars_as_right = self.__dict_to_bars(self.as_right, width=self.height, x=self.x, y=self.y, side=1,
                                                 anchor=self.anchor_right, nomenclature=self.nomenclature,
                                                 number_init=self.number_init_right) if self.as_right else []
        self.bars_as_inf = self.__dict_to_bars(self.as_inf, width=self.width, x=self.x, y=self.y, side=2,
                                               anchor=self.anchor_inf, nomenclature=self.nomenclature,
                                               number_init=self.number_init_inf) if self.as_inf else []
        self.bars_as_left = self.__dict_to_bars(self.as_left, width=self.height, x=self.x, y=self.y, side=3,
                                                anchor=self.anchor_left, nomenclature=self.nomenclature,
                                                number_init=self.number_init_left) if self.as_left else []
        self.all_bars = self.bars_as_sup + self.bars_as_right + self.bars_as_inf + self.bars_as_left

        self.all_elements = self.all_bars + self.stirrups

        # Beam attributes.
        if self.denomination is None:
            self.denomination = "Beam {0:.2f}x{1:.2f}".format(self.width, self.height)

    # Function that draws beam along longitudinal axe.
    def draw_longitudinal(self,
                          document: Drawing,
                          x: float = None,
                          y: float = None,
                          concrete_shape: bool = True,
                          bars: bool = True,
                          columns: bool = True,
                          columns_axes: bool = True,
                          stirrups: bool = True,
                          middle_axe: bool = True,
                          middle_axe_symbol: str = "A",
                          dim: bool = True,
                          dim_style: str = "EZ_M_25_H25_CM",
                          unifilar_bars: bool = False,
                          unifilar_stirrups: bool = True,
                          settings: dict = BEAM_SET_LONG) -> dict:
        """
        Draws the longitudinal section of the beam.

        :param document: The DXF document where the beam will be drawn.
        :type document: Drawing
        :param x: X-coordinate of the left bottom corner of beam concrete shape.
        :type x: float, optional
        :param y: Y-coordinate of the left bottom corner of beam concrete shape.
        :type y: float, optional
        :param concrete_shape: If True, the concrete shape of the beam is drawn.
        :type concrete_shape: bool
        :param bars: If True, the reinforcement bars are drawn.
        :type bars: bool
        :param columns: If True, the columns are drawn.
        :type columns: bool
        :param columns_axes: If True, the axes of the columns are drawn.
        :type columns_axes: bool
        :param stirrups: If True, the stirrups are drawn.
        :type stirrups: bool
        :param middle_axe: If True, the middle axis of the beam is drawn.
        :type middle_axe: bool
        :param middle_axe_symbol: Symbol to represent the middle axis.
        :type middle_axe_symbol: str
        :param dim: If True, dimensions are drawn.
        :type dim: bool
        :param dim_style: The dimension style to be used.
        :type dim_style: str
        :param unifilar_bars: If True, the reinforcement bars are drawn as unifilar.
        :type unifilar_bars: bool
        :param unifilar_stirrups: If True, the stirrups are drawn as unifilar.
        :type unifilar_stirrups: bool
        :param settings: Dictionary of drawing settings. Default is `BEAM_SET_LONG`.
        :type settings: dict

        :return: A list of graphical entities representing the longitudinal section of the beam.
        :rtype: list
        """
        if x is None:
            x = self.x
        if y is None:
            y = self.y

        elements = {}
        concrete_dict = None
        bar_dict_list = []
        stirrup_dict_list = []
        columns_elements = []
        columns_axes_elements = []
        dim_elements = []
        beam_axe_elements = []

        delimit_axe_height = self.height * 4
        delimit_axe_y = y - self.height

        # Drawing concrete shape.
        if concrete_shape:
            concrete_dict = self.concrete.draw_longitudinal(document=document,
                                                            x=x,
                                                            y=y,
                                                            dimensions=dim,
                                                            dimensions_inner=False,
                                                            settings=settings["concrete_settings"])

        # Drawing columns.
        if self.columns:
            for i, x_column in enumerate(self.columns_pos):
                x_column_relative = x + x_column
                x_delimit_axe_relative = x_column_relative + self.columns[i][0] / 2
                if columns:
                    columns_elements += rect(doc=document,
                                             width=self.columns[i][0],
                                             height=self.columns[i][1],
                                             x=x_column_relative,
                                             y=y,
                                             fill=True,
                                             polyline=True)
                if columns_axes:
                    columns_axes_elements += delimit_axe(document=document,
                                                         x=x_delimit_axe_relative,
                                                         y=delimit_axe_y,
                                                         height=delimit_axe_height,
                                                         radius=0.1,
                                                         text_height=0.1,
                                                         symbol=self.columns_symbol[i],
                                                         attr=GfxAttribs(linetype="CENTER"))

        # Drawing axe.
        if middle_axe:
            beam_axe_elements += delimit_axe(document=document,
                                             x=x - self.height,
                                             y=y + self.height / 2,
                                             height=self.length + self.height * 2,
                                             symbol=middle_axe_symbol,
                                             direction=Direction.HORIZONTAL,
                                             attr=GfxAttribs(linetype="CENTER"))
        # Drawing of stirrups.
        if self.stirrups:
            if stirrups:
                for stirrup in self.stirrups:
                    stirrup_dict_list.append(stirrup.draw_longitudinal(document=document,
                                                                       x=x + (stirrup.x - self.x),
                                                                       y=y + (stirrup.y - self.y),
                                                                       unifilar=unifilar_stirrups))
            # Drawing dimensions.
            if dim:
                dim_y = y + self.height * 2
                for stirrup in self.stirrups:
                    dim_elements += dim_linear(document=document,
                                               p_base=(
                                               x + (stirrup.x - self.x) + stirrup.reinforcement_length / 2, dim_y),
                                               p1=(x + (stirrup.x - self.x), dim_y),
                                               p2=(x + (stirrup.x - self.x) + stirrup.reinforcement_length, dim_y),
                                               dimstyle=dim_style)

        # Drawing of bars.
        if self.as_sup or self.as_right or self.as_inf or self.as_left:
            if bars:
                bar_dict_list.append(self.bars_as_sup[0].draw_longitudinal(document=document,
                                                                           x=x + (self.bars_as_sup[0].x - self.x),
                                                                           y=y + (self.bars_as_sup[0].y - self.y),
                                                                           unifilar=unifilar_bars,
                                                                           dimensions=False,
                                                                           denomination=False))  # Only mayor bar.
                bar_dict_list.append(self.bars_as_inf[0].draw_longitudinal(document=document,
                                                                           x=x + (self.bars_as_inf[0].x - self.x),
                                                                           y=y + (self.bars_as_inf[0].y - self.y),
                                                                           unifilar=unifilar_bars,
                                                                           dimensions=False,
                                                                           denomination=False))  # Only inferior mayor bar.
                for bar in self.bars_as_left:
                    bar_dict_list.append(bar.draw_longitudinal(document=document,
                                                               x=x + (bar.x - self.x),
                                                               y=y + (bar.y - self.y),
                                                               unifilar=unifilar_bars,
                                                               dimensions=False,
                                                               denomination=False))  # Only left bars.

        # Setting groups of elements in dictionary.
        elements["concrete"] = concrete_dict
        elements["bars"] = bar_dict_list
        elements["stirrups"] = stirrup_dict_list
        elements["columns_elements"] = columns_elements
        elements["columns_axes_elements"] = columns_axes_elements
        elements["dimensions_elements"] = dim_elements
        elements["beam_axe_elements"] = beam_axe_elements
        elements["all_elements"] = (elements["concrete"]["all_elements"] +
                                    list(chain(*[bar_dict["all_elements"] for bar_dict in elements["bars"]])) +
                                    list(chain(*[st_dict["all_elements"] for st_dict in elements["stirrups"]])) +
                                    columns_elements +
                                    columns_axes_elements +
                                    dim_elements +
                                    beam_axe_elements)

        return elements

    def draw_transverse(self,
                        document: Drawing,
                        x: float = None,
                        y: float = None,
                        x_section: float = None,
                        unifilar: bool = False,
                        dimensions: bool = True,
                        settings: dict = BEAM_SET_TRANSVERSE) -> dict:
        """
        Draws the transverse section of the beam at a given x-section.

        :param document: The DXF document where the beam will be drawn.
        :type document: Drawing
        :param x: X-coordinate of the starting point for drawing.
        :type x: float
        :param y: Y-coordinate of the starting point for drawing.
        :type y: float
        :param x_section: The x-coordinate of the section to be drawn.
        :type x_section: float, optional
        :param unifilar: If True, the bars are drawn as unifilar.
        :type unifilar: bool
        :param dimensions: If True, dimensions are drawn.
        :type dimensions: bool
        :param settings: Dict with beam transverse drawing settings.
        :type settings: dict

        :return: A dict of graphical entities representing the transverse section of the beam.
        :rtype: dict
        """
        if x is None:
            x = self.x
        if y is None:
            y = self.y

        if x_section is None:
            x_section = self.length / 2

        # Checking if x given is in beam length.
        if not 0 <= x_section <= self.length:
            return {}

        elements = {}
        bar_dict_list = []
        stirrup_dict_list = []

        # Obtaining elements for given x_section.
        entities = self.__elements_section(x=x_section)

        # Filtering elements.
        bars = filter(lambda e: e.element_type == ElementTypes.BAR, entities)
        stirrups = filter(lambda e: e.element_type == ElementTypes.STIRRUP, entities)

        # Concrete shape drawing.
        concrete_dict = self.concrete.draw_transverse(document=document,
                                                      x=x,
                                                      y=y,
                                                      dimensions=dimensions,
                                                      dimensions_boxing=True,
                                                      dimensions_inner=False,
                                                      settings=settings["concrete_settings"])

        # Drawing of bars.
        for bar in bars:
            bar_dict_list.append(bar.draw_transverse(document=document, x=x, y=y))

        # Drawing of stirrups.
        for stirrup in stirrups:
            delta_x = self.cover - max(self.max_db_sup, self.max_db_inf) / 2 - stirrup.diameter
            delta_y = self.cover - self.max_db_inf / 2 - stirrup.diameter

            stirrup_dict_list.append(stirrup.draw_transverse(document=document,
                                                             x=x + delta_x,
                                                             y=y + delta_y,
                                                             unifilar=unifilar))

        # Setting groups of elements in dictionary.
        elements["concrete"] = concrete_dict
        elements["bars"] = bar_dict_list
        elements["stirrups"] = stirrup_dict_list
        elements["all_elements"] = (elements["concrete"]["all_elements"] +
                                    list(chain(
                                        *[bar_dict["all_elements"] for bar_dict in elements["bars"]])) +
                                    list(chain(
                                        *[stirrup_dict["all_elements"] for stirrup_dict in elements["stirrups"]])))

        return elements

    # Function that draws the rebar detailing.
    def draw_longitudinal_rebar_detailing(self,
                                          document: Drawing,
                                          x: float = None,
                                          y: float = None,
                                          unifilar: bool = True,
                                          columns_axes: bool = True,
                                          settings: dict = BEAM_SET_LONG_REBAR) -> dict:
        """
        Draws the longitudinal rebar detailing for the beam.

        :param document: The DXF document where the detailing will be drawn.
        :type document: Drawing
        :param x: X-coordinate of the starting point for drawing.
        :type x: float, optional
        :param y: Y-coordinate of the starting point for drawing.
        :type y: float, optional
        :param unifilar: If True, the rebar is drawn as unifilar.
        :type unifilar: bool
        :param columns_axes: If True, the axes of the columns are drawn.
        :type columns_axes: bool
        :param settings: Dict with beam longitudinal rebar drawing settings.
        :type settings: dict

        :return: A list of graphical entities representing the longitudinal rebar detailing.
        :rtype: list
        """

        elements = {"text_elements": [],
                    "bars": [],
                    "barline_elements": [],
                    "columns_axes_elements": [],
                    "all_elements": []}

        def __draw_section_bars(bars: list = None,
                                spacing: float = 0.2,
                                rebar_x: float = None,
                                rebar_y: float = None,
                                barline=True,
                                text_reference: str = None) -> tuple:

            bars_elements = []
            text_elements = []
            barline_elements = []

            if text_reference:
                text_elements += text(document=document,
                                      text=text_reference,
                                      height=settings["text_height"],
                                      point=(rebar_x - self.height * 2 * 0.8, rebar_y + self.height * 0.05 - spacing))

            unique_bars_list = []
            for bar in bars:
                if bar.denomination not in unique_bars_list:
                    rebar_y -= spacing + bar.box_height
                    bars_elements.append(bar.draw_longitudinal(document=document,
                                                               x=rebar_x,
                                                               y=rebar_y,
                                                               unifilar=unifilar,
                                                               dimensions=True))
                    unique_bars_list.append(bar.denomination)
            rebar_y -= spacing

            if barline:
                barline_elements += rect(doc=document,
                                         width=self.length + self.height * 4,
                                         height=0,
                                         x=rebar_x - self.height * 2,
                                         y=rebar_y,
                                         sides=(1, 0, 0, 0))

            return text_elements, bars_elements, barline_elements, rebar_x, rebar_y

        rebar_x, rebar_y = x, y
        if self.as_sup:
            barline = True if self.as_right or self.bars_as_left or self.bars_as_inf else False
            elements_tuple = __draw_section_bars(bars=self.bars_as_sup,
                                                 spacing=0.2,
                                                 rebar_x=rebar_x,
                                                 rebar_y=rebar_y,
                                                 barline=barline,
                                                 text_reference="As sup.")

            text_elements, bars, barline_elements, rebar_x, rebar_y = elements_tuple

            elements["text_elements"].extend(text_elements)
            elements["bars"].extend(bars)
            elements["barline_elements"].extend(barline_elements)

        if self.as_left:
            barline = True if self.bars_as_left or self.bars_as_inf else False
            elements_tuple = __draw_section_bars(bars=self.bars_as_right,
                                                 spacing=0.2,
                                                 rebar_x=rebar_x,
                                                 rebar_y=rebar_y,
                                                 barline=barline,
                                                 text_reference="As right.")

            text_elements, bars, barline_elements, rebar_x, rebar_y = elements_tuple

            elements["text_elements"].extend(text_elements)
            elements["bars"].extend(bars)
            elements["barline_elements"].extend(barline_elements)

        if self.as_left:
            barline = True if self.bars_as_inf else False
            elements_tuple = __draw_section_bars(bars=self.bars_as_left,
                                                 spacing=0.2,
                                                 rebar_x=rebar_x,
                                                 rebar_y=rebar_y, barline=barline,
                                                 text_reference="As left.")

            text_elements, bars, barline_elements, rebar_x, rebar_y = elements_tuple

            elements["text_elements"].extend(text_elements)
            elements["bars"].extend(bars)
            elements["barline_elements"].extend(barline_elements)

        if self.as_inf:
            elements_tuple = __draw_section_bars(bars=self.bars_as_inf,
                                                 spacing=0.2,
                                                 rebar_x=rebar_x,
                                                 rebar_y=rebar_y,
                                                 barline=False,
                                                 text_reference="As inf.")

            text_elements, bars, barline_elements, rebar_x, rebar_y = elements_tuple

            elements["text_elements"].extend(text_elements)
            elements["bars"].extend(bars)
            elements["barline_elements"].extend(barline_elements)

        if columns_axes:
            for i, x_column in enumerate(self.columns_pos):
                delimit_axe_height = y - rebar_y
                delimit_axe_y = rebar_y
                x_column_relative = x + x_column
                x_delimit_axe_relative = x_column_relative + self.columns[i][0] / 2

                elements["columns_axes_elements"] += delimit_axe(document=document,
                                                                 x=x_delimit_axe_relative,
                                                                 y=delimit_axe_y,
                                                                 height=delimit_axe_height,
                                                                 radius=0.1,
                                                                 text_height=0.1,
                                                                 attr=GfxAttribs(linetype="CENTER"))

        # Setting groups of elements in dictionary.
        elements["all_elements"] = (elements["text_elements"] +
                                    list(chain(*[bar_dict["all_elements"] for bar_dict in elements["bars"]])) +
                                    elements["barline_elements"] +
                                    elements["columns_axes_elements"])

        return elements

    def draw_transverse_rebar_detailing(self, document: Drawing,
                                        x: float = None,
                                        y: float = None,
                                        x_section: float = None,
                                        unifilar: bool = False,
                                        dimensions: bool = True) -> dict:
        """
        Draws the transverse rebar detailing for the beam at a given x-section.

        :param document: The DXF document where the detailing will be drawn.
        :type document: Drawing
        :param x: X-coordinate of the starting point for drawing.
        :type x: float, optional
        :param y: Y-coordinate of the starting point for drawing.
        :type y: float, optional
        :param x_section: The x-coordinate of the section to be drawn.
        :type x_section: float, optional
        :param unifilar: If True, the rebar is drawn as unifilar.
        :type unifilar: bool
        :param dimensions: If True, dimensions are drawn.
        :type dimensions: bool

        :return: A dict of graphical entities representing the transverse rebar detailing.
        :rtype: dict
        """
        if x_section is None:
            x_section = self.length / 2

        elements = {"stirrups": []}

        # Checking if x given is in beam length.
        if not 0 <= x_section <= self.length:
            return []

        stirrups = self.__elements_section(elements=self.stirrups,
                                           x=x_section)

        for stirrup in stirrups:
            elements["stirrups"].append(stirrup.draw_transverse(document=document,
                                                                x=x,
                                                                y=y,
                                                                unifilar=unifilar,
                                                                dimensions=dimensions))

        # Setting groups of elements in dictionary.
        elements["all_elements"] = list(chain(*[stirrup_dict["all_elements"] for stirrup_dict in elements["stirrups"]]))

        return elements

    def __dict_to_bars(self, bars: dict,
                       width: float,
                       x: float,
                       y: float,
                       side: int,
                       anchor: list,
                       nomenclature: str = None,
                       number_init: int = None) -> list:
        """
        Converts a dictionary of bars into a list of Bar objects, placing them according to the specified side of
        the beam.

        :param bars: Dictionary containing bar information.
        :type bars: dict
        :param width: Width of the beam section.
        :type width: float
        :param x: X-coordinate of the starting point for placing the bars.
        :type x: float
        :param y: Y-coordinate of the starting point for placing the bars.
        :type y: float
        :param side: Indicates the side of the beam where the bars are placed (0: top, 1: right, 2: bottom, 3: left).
        :type side: int
        :param anchor: List of anchor lengths for each bar.
        :type anchor: list
        :param nomenclature: Nomenclature for the bars.
        :type nomenclature: str, optional
        :param number_init: Initial number for bar numbering.
        :type number_init: int, optional
        :return: A list of Bar objects representing the bars placed in the specified section of the beam.
        :rtype: list
        :raises ValueError: If the side is not within the range [0, 3].
        """
        if side < 0 or side > 3:
            raise ValueError

        list_bars, list_denom, list_pos, list_quantity = gen_symmetric_list(dictionary=bars,
                                                                            nomenclature=nomenclature,
                                                                            number_init=number_init,
                                                                            factor=1)  # Creating symetric list of bars.

        # Calculating separtion and definition of initial x_sep and y_sep.
        if side == 0 or side == 2:  # Sides top (0) or bottom (2).
            db_max = max(self.max_db_sup, self.max_db_inf) / 1
            width_util = width - self.cover * 2 + (db_max - max(list_bars))
            separation = width_util / (len(list_bars) - 1)
            x_sep, y_sep = -separation, 0

        else:  # Sides right (1) or left (3).
            db_max = max([self.max_db_right, self.max_db_left]) / 1
            width_util = width - self.cover * 2
            separation = width_util / (len(list_bars) + 1)
            x_sep, y_sep = 0, 0

        # Generating bars.
        entities, delta_x, delta_y, orientation = [], 0, 0, Orientation.BOTTOM
        for i, db in enumerate(list_bars):
            # Calculation of x_delta and y_delta.
            if side == 0 or side == 2:
                x_sep += separation
                delta_x = -self.cover + db_max / 2
                if side == 0:
                    delta_y = self.height - self.cover - 3 * db / 2 - anchor[i]
                    delta_y_transverse = self.height - self.cover + self.max_db_sup / 2 - db
                    orientation = Orientation.RIGHT
                if side == 2:
                    delta_y = self.cover - self.max_db_inf / 2
                    delta_y_transverse = self.cover - self.max_db_inf / 2
                    orientation = Orientation.TOP
            else:
                y_sep += separation
                delta_y = self.cover - db / 2

                if side == 1:
                    delta_x = self.cover + db - max(
                        [self.max_db_sup, self.max_db_inf, self.max_db_right]) / 2 - self.width
                    delta_y_transverse = delta_y
                    orientation = Orientation.RIGHT
                if side == 3:
                    delta_x = - self.cover + max([self.max_db_sup, self.max_db_inf, self.max_db_left]) / 2
                    delta_y_transverse = delta_y
                    orientation = Orientation.RIGHT

            x_bar_long = x + self.cover
            x_bar_transverse = x_sep - delta_x
            y_bar_long = y + y_sep + delta_y
            y_bar_transverse = y_sep + delta_y_transverse

            entities.append(Bar(reinforcement_length=self.length - self.cover * 2,
                                diameter=db,
                                x=x_bar_long,
                                y=y_bar_long,
                                left_anchor=anchor[i],
                                right_anchor=anchor[i],
                                mandrel_radius=db,
                                orientation=orientation,
                                transverse_center=(x_bar_transverse, y_bar_transverse),
                                denomination=list_denom[i],
                                position=list_pos[i],
                                quantity=list_quantity[i]))

        return entities

    def __elements_section(self,
                           elements: list = None,
                           x: float = None) -> list:
        """
        Filters and returns the elements that are located within the specified X-coordinate section of the beam.

        :param elements: List of elements (bars and stirrups) to filter.
        :type elements: list, optional
        :param x: X-coordinate of the beam section. If None, defaults to the middle of the beam.
        :type x: float, optional
        :return: List of elements located within the specified X-coordinate section.
        :rtype: list
        """
        if elements is None:
            elements = self.all_elements

        if x is None:
            x = self.x + self.length / 2  # Default value if is nothing entered.

        # Filtering elements by X coordinate.
        entities = []
        for element in elements:
            if element.x <= self.x + x <= element.x + element.reinforcement_length:
                entities.append(element)

        return entities

    def __list_to_stirrups(self) -> list:
        """
        Generates and returns a list of Stirrup objects based on data given at initialization.

        :return: A list of Stirrup objects representing the stirrups in the beam.
        :rtype: list
        """
        stirrup_width = self.width - (self.cover - max(self.max_db_sup, self.max_db_inf) / 2) * 2
        stirrup_height = self.height - (self.cover * 2 - self.max_db_sup / 2 - self.max_db_inf / 2)
        stirrups = zip(self.stirrups_db, self.stirrups_length, self.stirrups_sep, self.stirrups_anchor, self.stirrups_x)
        mandrel_radius_sup = self.max_db_sup / 2
        mandrel_radius_inf = self.max_db_inf / 2

        # Generating stirrups.
        entities = []  # Empty list that will contain stirrups elements.
        n = 1
        for stirrup in stirrups:
            stirrup_x = self.x + stirrup[4]  # Difining x coordinate.
            stirrup_y = self.y + self.cover - self.max_db_inf / 2 - stirrup[0]  # Subract of thick of stirrup.

            # Creating of stirrup.
            entities.append(Stirrup(width=stirrup_width + stirrup[0] * 2,
                                    height=stirrup_height + stirrup[0] * 2,
                                    diameter=stirrup[0],
                                    reinforcement_length=stirrup[1],
                                    spacing=stirrup[2],
                                    x=stirrup_x,
                                    y=stirrup_y,
                                    mandrel_radius_top=mandrel_radius_sup,
                                    mandrel_radius_bottom=mandrel_radius_inf,
                                    anchor=stirrup[3],
                                    position="{0}S{1}".format(self.nomenclature, n)))
            n += 1

        return entities

    def __dict_to_position_bars(self):
        positions = {}
        for bar in self.all_bars:
            if bar.denomination not in positions:
                positions[bar.denomination] = {"denomination:": bar.denomination, "quantity": 1, "bar": bar}
            else:
                positions[bar.denomination]["quantity"] += 1

    def draw_table_rebar_detailing(self,
                                   document: Drawing,
                                   x: float = None,
                                   y: float = None) -> dict:
        # Getting data.
        data = self.extract_data()

        # Creating table.
        table = CADTable(data=data, labels=["POSITION", "DIAMETER", "SPACING", "QUANTITY", "LENGTH", "TOTAL LENGTH",
                                            "WEIGHT"])

        # Drawing table.
        elements = table.draw_table(document=document, x=x, y=y)

        return elements

    def extract_data(self) -> list:
        data = []
        position_list = []
        for element in (self.all_bars + self.stirrups):
            if element.position not in position_list:
                position = element.position
                diameter = element.diameter
                spacing = "{0}".format(element.spacing) if element.element_type == ElementTypes.STIRRUP else "-"
                quantity = element.quantity if element.element_type == ElementTypes.STIRRUP else self.positions[element.position]["quantity"]
                length = "{0:.2f}".format(float(element.length))
                total_length = "{0:.2f}".format(element.quantity * element.length)

                weight = "{0:.2f}".format(element.quantity * element.weight)
                total_weight = 0

                data.append([position, diameter, spacing, quantity, length, total_length, weight])
                position_list.append(element.position)

        return data



# -*- coding: utf-8 -*-

# Local imports.

# External imports.
from enum import Enum
from math import cos, sin, pi


# Enums.
class Aligment(Enum):
    MTEXT_TOP_LEFT = 1
    MTEXT_TOP_CENTER = 2
    MTEXT_TOP_RIGHT = 3
    MTEXT_MIDDLE_LEFT = 4
    MTEXT_MIDDLE_CENTER = 5
    MTEXT_MIDDLE_RIGHT = 6
    MTEXT_BOTTOM_LEFT = 7
    MTEXT_BOTTOM_CENTER = 8
    MTEXT_BOTTOM_RIGHT = 9


class ColumnTypes(Enum):
    RECTANGULAR = 0
    CIRCULAR = 1


class Direction(Enum):
    HORIZONTAL = 0
    VERTICAL = 1


class DRotation(Enum):
    CLOCKWISE = 0
    COUNTERCLOCKWISE = 1


class ElementTypes(Enum):
    BAR = 0
    STIRRUP = 1
    BEAM = 2
    COLUMN = 3
    CONCRETE = 4
    CADTABLE = 21


class Orientation(Enum):
    TOP = 0
    RIGHT = 1
    BOTTOM = 2
    LEFT = 3


# Specific weights.
STEEL_WEIGHT = 7850  # kg / m3

CONCRETE_WEIGHT = 2400  # kg / m3

# General constants.
COS45 = cos(45 * pi / 180)
SIN45 = sin(45 * pi / 180)

# Settings dictionaries.
# Bar.
BAR_SET_LONG = {"text_dim_distance_horizontal": 0.05,
                "text_dim_distance_vertical": 0.05,
                "text_dim_height": 0.05,
                "text_denomination_distance": 0.1,
                "text_denomination_height": 0.05}

BAR_SET_TRANSVERSE = {"text_dim_distance_horizontal": 0.05,
                      "text_dim_distance_vertical": 0.05,
                      "text_dim_height": 0.05,
                      "text_denomination_distance": 0.1,
                      "text_denomination_height": 0.05}

# Beam.
BEAM_SET_LONG = {"concrete_settings": {"text_dim_distance_horizontal": 0.6,
                                       "text_dim_distance_vertical": 0.10,
                                       "dim_style_boxing": "EZ_M_25_H25_CM",
                                       "dim_style_inner": "EZ_M_10_H25_CM"},
                 "text_dim_horizontal_distance": 0.05,
                 "text_dim_height": 0.05}

BEAM_SET_TRANSVERSE = {"concrete_settings": {"text_dim_distance_horizontal": 0.10,
                                             "text_dim_distance_vertical": 0.10,
                                             "text_dim_inner_perpendicular_distance": 0.05,
                                             "dim_style_boxing": "EZ_M_10_H25_CM",
                                             "dim_style_inner": "EZ_M_10_H25_CM"},
                       "text_dim_distance": 0.05,
                       "text_dim_height": 0.05}

BEAM_SET_LONG_REBAR = {"text_height": 0.05}

# CADTable.
CADTABLE_SET_DEFAULT = {"content_row_height": 0.3,
                        "content_column_width": 0.6,
                        "content_text_height": 0.10,
                        "content_fit": True,
                        "labels_row_height": 0.3,
                        "labels_column_width": 1,
                        "labels_text_height": 0.2,
                        "labels_fit": True}

# Column.
COLUMN_SET_LONG = {"concrete_settings": {"text_dim_distance_horizontal": 0.15,
                                         "text_dim_distance_vertical": 0.25,
                                         "dim_style_boxing": "EZ_M_25_H25_CM",
                                         "dim_style_inner": "EZ_M_10_H25_CM"},
                   "text_dim_horizontal_distance": 0.05,
                   "text_dim_height": 0.05}

COLUMN_SET_TRANSVERSE = {"concrete_settings": {"text_dim_distance_horizontal": 0.05,
                                               "text_dim_distance_vertical": 0.05,
                                               "dim_style_boxing": "EZ_M_10_H25_CM",
                                               "dim_style_inner": "EZ_M_10_H25_CM"},
                         "text_dim_height": 0.05}

COLUMN_SET_LONG_REBAR = {"bar_settings": {"text_dim_distance_horizontal": 0.05,
                                          "text_dim_distance_vertical": 0.05,
                                          "text_dim_height": 0.05,
                                          "text_denomination_distance": 0.05,
                                          "text_denomination_height": 0.05},
                         "text_height": 0.05,
                         "spacing": 0.3}

COLUMN_SET_TRANSVERSE_REBAR = {"bar_settings": {"text_dim_distance_horizontal": 0.05,
                                                "text_dim_distance_vertical": 0.05,
                                                "text_dim_height": 0.05,
                                                "text_denomination_distance": 0.05,
                                                "text_denomination_height": 0.05},
                               "text_height": 0.05,
                               "spacing": 0.3}

# Concrete.
CONCRETE_SET_LONG = {"text_dim_distance_horizontal": 0.25,
                     "text_dim_distance_vertical": 0.25,
                     "dim_style_boxing": "EZ_M_10_H25_CM",
                     "dim_style_inner": "EZ_M_10_H25_CM"}

CONCRETE_SET_TRANSVERSE = {"text_dim_distance_horizontal": 0.25,
                           "text_dim_distance_vertical": 0.25,
                           "text_dim_inner_perpendicular_distance": 0.05,
                           "dim_style_boxing": "EZ_M_10_H25_CM",
                           "dim_style_inner": "EZ_M_10_H25_CM"}

# Stirrups.
STIRRUP_SET_TRANSVERSE = {"text_dim_distance_horizontal": 0.05,
                          "text_dim_distance_vertical": 0.10,
                          "text_dim_distance_anchor": 0.10,
                          "text_distance_length_count": 0.1,
                          "text_length_count_height": 0.05,
                          "dim_style": "EZ_M_10_H25_CM"}
